# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test integer transform exceptions."""

# pylint: disable=unidiomatic-typecheck

import pytest

from baseline import Baseline

from plum.bigendian import uint8
from plum.conformance import wrap_message
from plum.exceptions import PackError
from plum.utilities import pack


class TestPack:

    """Test pack transform exceptions."""

    def test_not_int_val(self):
        """Test not integer values."""
        with pytest.raises(PackError) as trap:
            pack("str", uint8)

        expected = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            |        | 'str' |       | uint8  |
            +--------+-------+-------+--------+

            TypeError occurred during pack operation:

            value type <class 'str'> not int-like (no to_bytes() method)
            """
        )

        assert wrap_message(trap.value) == expected

    def test_less_than_minimum(self):
        """Test value too little."""
        with pytest.raises(PackError) as trap:
            pack(-1, uint8)

        expected_context = Baseline(
            """
            can't convert negative int to unsigned
            """
        )

        expected = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            |        | -1    |       | uint8  |
            +--------+-------+-------+--------+

            OverflowError occurred during pack operation:

            can't convert negative int to unsigned
            """
        )

        context_message_match = wrap_message(trap.value.__context__) == expected_context
        message_match = wrap_message(trap.value) == expected
        assert context_message_match and message_match
        assert type(trap.value.__context__) is OverflowError

    def test_exceeds_maximum(self):
        """Test value too large."""
        with pytest.raises(PackError) as trap:
            pack(256, uint8)

        expected_context = Baseline(
            """
            int too big to convert
            """
        )

        expected = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            |        | 256   |       | uint8  |
            +--------+-------+-------+--------+

            OverflowError occurred during pack operation:

            int too big to convert
            """
        )

        context_message_match = wrap_message(trap.value.__context__) == expected_context
        message_match = wrap_message(trap.value) == expected
        assert context_message_match and message_match
        assert type(trap.value.__context__) is OverflowError



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test base view pack methods."""

from baseline import Baseline

from plum.littleendian import uint16


class TestPack:

    """Test base view pack methods."""

    EXPECT_DUMP = Baseline(
        """
        +--------+-------+-------+--------+
        | Offset | Value | Bytes | Format |
        +--------+-------+-------+--------+
        | 1      | 513   | 01 02 | uint16 |
        +--------+-------+-------+--------+
        """
    )

    def test_pack(self):
        """Test base view pack() method."""
        buffer = bytearray([0, 1, 2, 3])
        view_16 = uint16.view(buffer, offset=1)
        assert view_16.ipack() == b"\x01\x02"

    def test_pack_and_dump(self):
        """Test base view pack_and_dump() method."""
        buffer = bytearray([0, 1, 2, 3])
        view_16 = uint16.view(buffer, offset=1)
        membytes, dump = view_16.ipack_and_dump()
        assert membytes == b"\x01\x02"
        assert str(dump) == self.EXPECT_DUMP



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Default sentinal."""


class NoDefault:

    """Default sentinal."""

    def __new__(cls):
        try:
            instance = NO_DEFAULT
        except NameError:
            instance = super().__new__(cls)

        return instance

    def __repr__(self):
        return "NO_DEFAULT"


NO_DEFAULT = NoDefault()



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Classes and utilities for packing/unpacking bytes."""

import inspect
import os
from typing import Any, List, Optional, Tuple, Type, TypeVar, Union

from . import dump as _dump
from . import exceptions


class DataMeta(type):

    """Meta class for plum types."""

    __nbytes__: Union[int, None] = None
    __implementation__: Union[str, Tuple[str, int, Tuple[str, ...]]]
    __format_name__: Optional[str] = None

    @property
    def __hint__(cls) -> str:
        """Type hint."""
        return cls.__name__

    @property
    def name(cls) -> str:
        """Transform format name (for dump "Format" column)."""
        name = cls.__format_name__

        if name is None:
            name = cls.__name__

        return name

    @property
    def nbytes(cls) -> int:
        """Transform format size in bytes."""
        nbytes = cls.__nbytes__

        if nbytes is None:
            raise exceptions.SizeError(f"{cls.__name__!r} instance sizes vary")

        return nbytes

    @property
    def implementation(cls) -> str:
        """Generated implementation code."""
        code = cls.__implementation__

        if isinstance(code, tuple):  # pragma: no cover
            code = ""

        return code

    @implementation.setter
    def implementation(cls, value: str) -> None:  # pragma: no cover
        # pylint: disable=no-self-use
        frame = list(inspect.stack())[1].frame
        namespace = frame.f_locals
        path = os.path.abspath(frame.f_code.co_filename)
        selections = (
            () if value == "all" else tuple(v.strip() for v in value.split(","))
        )
        namespace["__implementation__"] = (path, frame.f_lineno, selections)

    def __pack__(
        cls, value: Any, pieces: List[bytes], dump: Optional[_dump.Record] = None
    ) -> None:
        raise TypeError(f"{cls.__name__!r} does not support pack")  # pragma: no cover

    def __unpack__(
        cls, buffer: bytes, offset: int, dump: Optional[_dump.Record] = None
    ) -> Tuple[Any, int]:
        raise TypeError(f"{cls.__name__!r} does not support unpack")  # pragma: no cover

    def __str__(cls) -> str:
        return f"<transform class {cls.name!r}>"

    def __repr__(cls) -> str:
        return f"<transform class {cls.name!r}>"


CLS = TypeVar("CLS", bound="Data")


class Data(metaclass=DataMeta):

    """Packable/Unpacked bytes base class."""

    __nbytes__: Union[int, None] = None
    __format_name__: str

    @property
    def nbytes(self) -> int:
        """Transform format size in bytes."""
        nbytes = self.__nbytes__

        if nbytes is None:
            nbytes = len(self.ipack())

        return nbytes

    @classmethod
    def __pack__(
        cls: Type[CLS],
        value: Any,
        pieces: List[bytes],
        dump: Optional[_dump.Record] = None,
    ) -> None:
        raise TypeError(f"{cls.__name__!r} does not support pack")  # pragma: no cover

    @classmethod
    def __unpack__(
        cls: Type[CLS], buffer: bytes, offset: int, dump: Optional[_dump.Record] = None
    ) -> Tuple[CLS, int]:
        # pylint: disable=unused-argument
        raise TypeError(f"{cls.__name__!r} does not support unpack")  # pragma: no cover

    @classmethod
    def __view__(cls, buffer, offset=0):
        """Create view of bytes buffer."""
        raise TypeError(f"{cls.__name__!r} does not support view()")

    @property
    def dump(self) -> _dump.Dump:
        """Summary containing details of bytes and layout."""
        _buffer, dump = self.ipack_and_dump()
        return dump

    @classmethod
    def pack(cls, value: Any) -> bytes:
        """Pack value as formatted bytes.

        :raises: ``PackError`` if type error, value error, etc.

        """
        pieces: List[bytes] = []
        try:
            # None -> dump
            cls.__pack__(value, pieces, None)
        except Exception as exc:
            # do it over to include dump in exception message
            cls.pack_and_dump(value)

            raise exceptions.ImplementationError() from exc  # pragma: no cover

        return b"".join(pieces)

    def ipack(self) -> bytes:
        """Pack instance as bytes.

        :raises: ``PackError`` if type error, value error, etc.

        """
        pieces: List[bytes] = []
        try:
            # None -> dump
            self.__pack__(self, pieces, None)
        except Exception as exc:
            # do it over to include dump in exception message
            self.ipack_and_dump()

            raise exceptions.ImplementationError() from exc  # pragma: no cover

        return b"".join(pieces)

    @classmethod
    def pack_and_dump(cls, value: Any) -> Tuple[bytes, _dump.Dump]:
        """Pack value as formatted bytes and produce bytes summary.

        :raises: ``PackError`` if type error, value error, etc.

        """
        dump = _dump.Dump()
        pieces: List[bytes] = []
        try:
            cls.__pack__(value, pieces, dump.add_record(fmt=cls.name))
        except Exception as exc:
            raise exceptions.PackError(dump=dump, exception=exc) from exc

        return b"".join(pieces), dump

    def ipack_and_dump(self) -> Tuple[bytes, _dump.Dump]:
        """Pack instance as bytes and produce bytes summary.

        :raises: ``PackError`` if type error, value error, etc.

        """
        dmp = _dump.Dump()
        pieces: List[bytes] = []
        try:
            self.__pack__(self, pieces, dmp.add_record(fmt=type(self).name))
        except Exception as exc:
            raise exceptions.PackError(dump=dmp, exception=exc) from exc

        return b"".join(pieces), dmp

    @classmethod
    def unpack(cls, buffer: bytes) -> Any:
        """Unpack value from formatted bytes.

        :raises: ``UnpackError`` if insufficient bytes, excess bytes, or value error

        """
        try:
            # None -> dump
            item, offset = cls.__unpack__(buffer, 0, None)
            if buffer[offset:]:
                raise exceptions.ExcessMemoryError()
            return item

        except Exception as exc:
            # do it over to include dump in exception message
            cls.unpack_and_dump(buffer)
            raise exceptions.ImplementationError() from exc  # pragma: no cover

    @classmethod
    def unpack_and_dump(cls, buffer: bytes) -> Tuple[Any, _dump.Dump]:
        r"""Unpack value from bytes and produce packed bytes summary.

        :raises: ``UnpackError`` if insufficient bytes, excess bytes, or value error

        """
        dmp = _dump.Dump()
        try:
            item, offset = cls.__unpack__(buffer, 0, dmp.add_record(fmt=cls.name))

            extra_bytes = buffer[offset:]

            if extra_bytes:
                for i in range(0, len(extra_bytes), 16):
                    record = dmp.add_record(memory=extra_bytes[i : i + 16])
                    if not i:
                        record.separate = True
                        record.value = "<excess bytes>"

                raise exceptions.ExcessMemoryError(extra_bytes)

        except Exception as exc:
            raise exceptions.UnpackError(dmp, exc) from exc

        return item, dmp

    @classmethod
    def view(cls, buffer: bytearray, offset: int = 0) -> Any:
        """Create view of bytes buffer."""
        return cls.__view__(buffer, offset)


# for easy injection of boost versions of utilities into this namespace
plum_namespace = globals()
packmethod_doc = Data.ipack.__doc__  # pylint: disable=no-member
unpackmethod_doc = Data.unpack.__doc__  # pylint: disable=no-member



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Bit field structure member definition."""

# pylint: disable=too-many-arguments

from dataclasses import dataclass
from enum import EnumMeta
from typing import Any, Dict, Optional, Tuple

from .._code_injector import CodeInjector
from .._default import NO_DEFAULT
from .._getbytes import getbytes
from .._typing import ByteOrderHint
from ..bitfields import BitFields, BitFieldsMeta
from ..dump import Record
from ..enum import EnumX
from ..flag import FlagX
from ..exceptions import InsufficientMemoryError
from ..none import NoneX
from .member import Member


@dataclass
class BitFieldsStore:

    """Temporary information store for generating implementation code."""

    is_first: bool = False
    is_last: bool = False
    group_name: str = ""
    nbytes: Optional[int] = None
    byteorder: str = "little"


class BitFieldMember(Member):

    """Bit field structure member definition.

    :param doc: accessor documentation string
    :param size: bit field width (in bits)
    :param lsb: least significant bit position
    :param typ: field type (e.g. ``IntEnum`` subclass), default uses type annotation
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param compute: initializer defaults to compute based on another member
    :param nbytes: bytes occupied by this bit field and ones that follow (0 => calculate)
    :param argrepr: format to represent member argument in structure repr

    """

    aux_store: BitFieldsStore

    _group_name: str
    _typ: Any

    _attributes_to_copy: Tuple[str, ...] = Member._attributes_to_copy + (
        "_typ",
        "_size",
        "_lsb",
        "_signed",
        "_group_name",
    )

    def __init__(
        self,
        doc: str,
        size: int,
        lsb: Optional[int],
        typ: type,
        signed: bool,
        default: Any,
        ignore: bool,
        readonly: bool,
        compute: bool,
        nbytes: int,
        argrepr: Optional[str],
    ) -> None:
        if lsb is not None:
            lsb = int(lsb)
            if lsb < 0:
                raise TypeError(
                    "bit field position must be greater than or equal to zero"
                )

        if signed and size < 2:
            raise ValueError("'size' must be 2 or greater for signed bit field")

        if size < 1:
            raise ValueError("'size' must be 1 or greater for unsigned bit field")

        super().__init__(
            doc,
            default,
            ignore,
            readonly,
            compute,
            argrepr,
            fmt=NoneX(),
            fmt_arg=None,
        )

        self._set_typ(typ)
        self._size = size
        self._lsb = lsb
        self._nbytes = nbytes
        self._signed = signed
        self._group_name = ""

    @property
    def lsb(self) -> int:
        """Bit field bit position."""
        return -1 if self._lsb is None else self._lsb

    @property
    def nbytes(self) -> int:
        """Number of bytes allocated for this and following bit fields.

        Return zero for bit fields included in previous bit field allocation.

        """
        if self._nbytes is None:  # pragma: no cover
            raise RuntimeError("internal error")
        return self._nbytes

    @property
    def signed(self) -> bool:
        """Bit field is a signed integer."""
        return self._signed

    @property
    def size(self) -> int:
        """Bit field width (in bits)."""
        return self._size

    @property
    def typ(self) -> type:
        """Field type (e.g. ``IntEnum`` subclass), default uses type annotation."""
        return self._typ

    def _set_typ(self, typ: type):
        # FUTURE - eliminate/conditionalize this code for performance
        try:
            is_int_subclass = issubclass(typ, (int, BitFields))

        except TypeError:
            is_int_subclass = False  # must not be class

        if isinstance(typ, (FlagX, EnumX)) or is_int_subclass:
            pass

        else:
            raise TypeError("bit field type must be int-like")

        self._typ = typ

    @property
    def argrepr(self) -> str:
        """Format string to represent member argument in structure repr."""
        if isinstance(self._argrepr, str):
            repr_format = self._argrepr

        elif isinstance(self._typ, EnumMeta):
            repr_format = (
                f"{self.name}={{repr(self.{self.name}).split(':', 1)[0].lstrip('<')}}"
            )

        else:
            repr_format = f"{self.name}={{self.{self.name}!r}}"

        return repr_format

    @classmethod
    def organize(cls, members: Dict[str, Member], byteorder: str, fieldorder: str):
        """Organize bit field members into groups."""
        collection = []
        group = {}

        for name, member in members.items():
            if isinstance(member, BitFieldMember) and not member.nbytes:
                group[name] = member
                continue

            if group:
                collection.append(group)
                group = {}

            if isinstance(member, BitFieldMember):
                group[name] = member

        if group:
            collection.append(group)

        for i, group in enumerate(collection):
            nbytes = cls.position_and_size(f"fields_{i}", group, fieldorder)
            for member in group.values():
                member.aux_store = BitFieldsStore(
                    byteorder=byteorder, nbytes=nbytes, group_name=f"bitfields_{i}"
                )
            list(group.values())[0].aux_store.is_first = True
            list(group.values())[-1].aux_store.is_last = True

    @staticmethod
    def position_and_size(
        group_name: str, members: Dict[str, "BitFieldMember"], fieldorder: str
    ) -> int:
        """Fill in any unassigned positions, determine overall size."""
        # fill in missing positions (and assign group name)
        lsb = 0
        field_member_properties = (
            members.values()
            if fieldorder == "least_to_most"
            else reversed(list(members.values()))
        )
        for field_member in field_member_properties:
            # pylint: disable=protected-access
            field_member._group_name = group_name
            if field_member._lsb is None:
                field_member._lsb = lsb
            lsb = field_member.lsb + field_member.size

        # check for overlap
        claimed_bits: Dict[int, str] = {}
        for name, member in members.items():
            for i in range(member.lsb, member.lsb + member.size):
                if i in claimed_bits:
                    raise TypeError(
                        f"bit field {name!r} overlaps with {claimed_bits[i]!r}"
                    )

                claimed_bits[i] = name

        # default/validate 'nbytes'
        numbits = max(member.lsb + member.size for member in members.values())
        nbytes_needed = (numbits + 7) // 8

        first_bitfield_name, first_bitfield = list(members.items())[0]
        nbytes = first_bitfield.nbytes

        if not nbytes:
            nbytes = first_bitfield._nbytes = nbytes_needed

        elif nbytes_needed > nbytes:
            raise TypeError(
                f"'nbytes' must be at least {nbytes_needed} for bitfield "
                f"{first_bitfield_name!r} (to have sufficient room for it "
                f"and those that follow)"
            )

        return nbytes

    @staticmethod
    def __unpack__(
        buffer: bytes,
        offset: int,
        dump: Optional[Record],
        byteorder: ByteOrderHint,
        nbytes: int,
        signed: bool,
    ) -> Tuple[int, int]:
        if dump is None:
            end = offset + nbytes

            if len(buffer) < end:
                raise InsufficientMemoryError(
                    "too few bytes to unpack, retry with dump to follow"
                )

            value = int.from_bytes(buffer[offset:end], byteorder, signed=signed)

            return value, end

        chunk, offset = getbytes(buffer, offset, dump, nbytes)

        value = int.from_bytes(chunk, byteorder, signed=signed)

        dump.value = value

        return value, offset

    def add_name_index(
        self,
        name: str,
        index: int,
        code_injector: CodeInjector,
    ) -> int:
        """Assign name, index number, and prepare for code generation."""
        # pylint: disable=too-many-statements, too-many-branches, too-many-locals
        index = super().add_name_index(name, index, code_injector)

        store = self.temp_store
        aux = self.aux_store

        # Erase normal member packing lines (from super)
        store.unpack = []
        store.unpack_and_dump = []
        store.pack = []
        store.pack_and_dump = []

        lo_mask = (1 << self.size) - 1
        sign_bit = 1 << (self.size - 1)

        cls_name = code_injector.get_expression(self.typ, f"cls.{self.name}.typ")

        dump_name = f"{aux.group_name}_dump"
        dump_group = [f"{dump_name} = dump.add_record()"]

        if isinstance(self.typ, BitFieldsMeta):
            dump_member = [
                # line 1
                f"dump_m_{self.name} = dump.add_record("
                f'access="{name}", value=int(m_{self.name}), fmt={cls_name})',
                # line 2
                f"{cls_name}.__add_bitfields_to_dump__("
                f"m_{self.name}, dump_m_{self.name}, bitoffset={self.lsb})",
            ]
        else:
            dump_member = [
                f"dump.add_record("
                f'access="{name}", bits={self.lsb, self.size}, '
                f"value=m_{self.name}, fmt={cls_name})"
            ]

        if aux.is_first:
            store.unpack += [
                f"{aux.group_name}, offset = "
                f"cls.{name}.__unpack__(buffer, offset, dump, "
                f'"{aux.byteorder}", nbytes={aux.nbytes}, signed={self.signed})',
                "",
            ]

            store.unpack_and_dump += dump_group + [
                f"{aux.group_name}, offset = "
                f"cls.{name}.__unpack__(buffer, offset, {dump_name}, "
                f'"{aux.byteorder}", nbytes={aux.nbytes}, signed={self.signed})',
                "",
            ]

        lines = [f"m_{name} = ({aux.group_name} >> {self.lsb}) & 0x{lo_mask:x}"]

        if self.signed:
            lines += [
                f"if m_{name} & 0x{sign_bit:x}:",
                f"    m_{name} -= 0x{1 << self.size:x}",
            ]

        if self.typ is not int:
            if isinstance(self.typ, BitFieldsMeta):
                lines.append(f"m_{name} = {cls_name}.from_int(m_{name})")
            else:
                lines.append(f"m_{name} = {cls_name}(m_{name})")

        store.unpack += lines
        store.unpack_and_dump += lines + dump_member

        if aux.is_first:
            lines = [f"{aux.group_name} = 0", ""]
            store.pack += lines
            store.pack_and_dump += dump_group + lines

        assert self.lsb is not None, "internal error"
        hi_mask = ~(lo_mask << self.lsb)
        member_name = f"m_{self.name}"
        if self.typ is bool:
            # don't let an integer > 1 be masked off and become False
            member_name = f"int(bool({member_name}))"
        op2 = f"({member_name} & 0x{lo_mask:x})"
        if self.lsb:
            op2 = f"({op2} << {self.lsb})"
        lines = [f"{aux.group_name} |= ({aux.group_name} & {hi_mask}) | {op2}"]
        store.pack += lines
        store.pack_and_dump += dump_member + lines

        if aux.is_last:
            store.pack += [
                "",
                f"pieces.append({aux.group_name}.to_bytes("
                f'{aux.nbytes}, "{aux.byteorder}", signed={self.signed}))',
            ]
            store.pack_and_dump += [
                "",
                f"pieces.append({aux.group_name}.to_bytes("
                f'{aux.nbytes}, "{aux.byteorder}", signed={self.signed}))',
                f"{dump_name}.value = str({aux.group_name})",
                f"{dump_name}.memory = pieces[-1]",
            ]

        if self.signed:
            neg_limit = -(1 << (self.size - 1))
            maxvalue = f"{-(1 + neg_limit)}"
            minvalue = f"-{-neg_limit}"
        else:
            minvalue = "0"
            maxvalue = f"{(1 << self.size) - 1}"

        if self.fset is None and not self.readonly:
            # override lines put in by super()
            if isinstance(self.typ, EnumMeta):
                store.setter = (
                    []
                )  # no need for range check, enum instantiation checks validity
            else:
                store.setter = [
                    f"if not {minvalue} <= value <= {maxvalue}:",
                    f'   raise ValueError("out of range, {minvalue} <= value <= {maxvalue}")',
                ]
            if self.typ is int:
                store.setter.append(f"self[{self.index}] = value")
            else:
                typ = cls_name.replace("cls.", "type(self).")
                store.setter.append(f"self[{self.index}] = {typ}(value)")

        if not self.default_is_a_factory:
            store.init_begin += [
                f"if not {minvalue} <= {self.name} <= {maxvalue}:",
                f'   raise ValueError("{self.name!r} out of range, {minvalue} <= {self.name} <= {maxvalue}")',
            ]

        return index


def bitfield_member(
    doc: str = "",
    *,
    size: int,
    lsb: Optional[int] = None,
    typ: type = int,
    signed: bool = False,
    default: Any = NO_DEFAULT,
    ignore: bool = False,
    readonly: bool = False,
    compute: bool = False,
    nbytes: int = 0,
    argrepr: Optional[str] = None,
) -> Any:
    """Bit field structure member definition.

    :param doc: accessor documentation string
    :param size: bit field width (in bits)
    :param lsb: least significant bit position
    :param signed: interpret as signed integer
    :param typ: field type (e.g. ``IntEnum`` subclass)
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param compute: initializer defaults to compute based on another member
    :param nbytes: bytes occupied by this bit field and ones that follow (0 => calculate)
    :param argrepr: format to represent member argument in structure repr

    """
    return BitFieldMember(
        doc, size, lsb, typ, signed, default, ignore, readonly, compute, nbytes, argrepr
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Structure type features."""

from enum import IntEnum

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.enum import EnumX
from plum.exceptions import PackError
from plum.littleendian import uint8, uint16
from plum.structure import Structure, bitfield_member, member
from plum.utilities import pack


class Custom(Structure):

    """Sample structure type."""

    mbr1: int = member(fmt=uint8)
    mbr2: int = member(fmt=uint16, default=0x9988)


sample_dump = Baseline(
    """
    +--------+--------+-------+-------+--------------------+
    | Offset | Access | Value | Bytes | Format             |
    +--------+--------+-------+-------+--------------------+
    |        |        |       |       | Custom (Structure) |
    | 0      | mbr1   | 1     | 01    | uint8              |
    | 1      | mbr2   | 2     | 02 00 | uint16             |
    +--------+--------+-------+-------+--------------------+
    """
)


class TestInit:

    """Test initializer."""

    @staticmethod
    def test_init_keyword():
        """Test initialization via keyword arguments (no defaults)."""
        custom = Custom(mbr1=1, mbr2=2)
        assert str(custom.dump) == sample_dump

    @staticmethod
    def test_init_keyword_default():
        """Test initialization via keyword arguments (use defaults)."""
        custom = Custom(mbr1=1)

        expected_dump = Baseline(
            """
            +--------+--------+-------+-------+--------------------+
            | Offset | Access | Value | Bytes | Format             |
            +--------+--------+-------+-------+--------------------+
            |        |        |       |       | Custom (Structure) |
            | 0      | mbr1   | 1     | 01    | uint8              |
            | 1      | mbr2   | 39304 | 88 99 | uint16             |
            +--------+--------+-------+-------+--------------------+
            """
        )

        assert str(custom.dump) == expected_dump


class TestIndexAccess:

    """Test __setitem__ and __getitem__ usages."""

    sample = Custom(mbr1=1, mbr2=2)
    sample_dump = sample_dump

    def test_valid_lookup(self):
        """Test __getitem__ with valid index."""
        assert self.sample[0] == 1
        assert self.sample[1] == 2

    def test_valid_set(self):
        """Test __setitem__ with valid index."""
        cls = type(self.sample)
        sample = cls(mbr1=0, mbr2=2)
        sample[0] = 1
        assert str(sample.dump) == self.sample_dump

    def test_valid_slice_set(self):
        """Test __setitem__ with valid slice index."""
        cls = type(self.sample)
        sample = cls(mbr1=11, mbr2=22)
        sample[0:2] = [1, 2]
        assert str(sample.dump) == self.sample_dump

    def test_invalid_lookup(self):
        """Test __getitem__ with out of range index."""
        with pytest.raises(IndexError) as trap:
            # pylint: disable=pointless-statement
            self.sample[2]

        assert str(trap.value) == "list index out of range"

    def test_invalid_set_index(self):
        """Test __setitem__ with out of range index."""
        with pytest.raises(IndexError) as trap:
            self.sample[2] = 2

        assert str(trap.value) == "list assignment index out of range"


class TestNameAccess:

    """Test member access via attribute."""

    sample = Custom(mbr1=1, mbr2=2)
    sample_dump = sample_dump

    # pylint: disable=unneeded-not

    def test_valid_lookup(self):
        """Test member get via attribute access with valid name."""
        assert self.sample.mbr1 == 1
        assert self.sample.mbr2 == 2

    def test_valid_set(self):
        """Test member set via attribute access with valid name."""
        sample = Custom(mbr1=0, mbr2=2)
        sample.mbr1 = 1
        assert str(sample.dump) == self.sample_dump

    def test_invalid_lookup(self):
        """Test member get via attribute access with invalid name."""
        with pytest.raises(AttributeError) as trap:
            # pylint: disable=no-member,pointless-statement
            self.sample.m3

        clsname = type(self.sample).__name__
        assert str(trap.value) == f"{clsname!r} object has no attribute 'm3'"

    def test_invalid_set(self):
        """Test member set via attribute access with valid name."""
        cls = type(self.sample)
        sample = cls(mbr1=0, mbr2=2)
        with pytest.raises(AttributeError) as trap:
            sample.m3 = 0

        assert str(trap.value) == f"{cls.__name__!r} object has no attribute 'm3'"


class Sample(Structure):

    """Sample structure type."""

    mbr1: int = member(fmt=uint8)
    mbr2: int = member(fmt=uint8)


class TestExceptions:

    """Test exception corner cases."""

    def test_member_duplicate(self):
        """Test member definition may only be used once in one structure class."""

        expected = Baseline(
            """
            invalid structure member 'xyz' definition, member instance can not be
            shared, create a new instance
            """
        )

        with pytest.raises(TypeError) as trap:

            class Struct(Structure):  # pylint: disable=unused-variable

                abc = member(fmt=uint8)
                xyz = abc

        assert wrap_message(trap.value) == expected

    def test_member_already_in_use(self):
        """Test member definition may only be used in one structure class."""

        expected = Baseline(
            """
            invalid structure member 'xyz' definition, member instance can not be
            shared, create a new instance
            """
        )

        memb = member(fmt=uint8)

        class FirstUse(Structure):  # pylint: disable=unused-variable

            xyz = memb

        with pytest.raises(TypeError) as trap:

            class SecondUse(Structure):  # pylint: disable=unused-variable

                xyz = memb

        assert wrap_message(trap.value) == expected

    def test_invalid_fieldorder(self):
        expected = Baseline(
            """
            fieldorder must be either "least_to_most" or "most_to_least"
            """
        )

        with pytest.raises(ValueError) as trap:

            class Struct(
                Structure, fieldorder="invalid"
            ):  # pylint: disable=unused-variable
                pass

        assert wrap_message(trap.value) == expected

    def test_bitfield_already_in_use(self):
        """Test bitfield member definition may only be used in one structure class."""

        expected = Baseline(
            """
            invalid structure member 'xyz' definition, member instance can not be
            shared, create a new instance
            """
        )

        bitfield = bitfield_member(typ=int, lsb=0, size=1)

        class FirstUse(Structure):  # pylint: disable=unused-variable

            xyz = bitfield

        with pytest.raises(TypeError) as trap:

            class SecondUse(Structure):  # pylint: disable=unused-variable

                xyz = bitfield

        assert wrap_message(trap.value) == expected

    def test_invalid_member_type(self):
        """Test member type not a plum type."""
        with pytest.raises(TypeError) as trap:

            class Bad(Structure):  # pylint: disable=unused-variable
                """Sample structure with member with invalid type."""

                m1: int = member(fmt=int)

        exp_message = Baseline(
            """
            'fmt' must be a data store, transform, or a factory function
            """
        )

        assert str(trap.value) == exp_message

    def test_defined_struct_pack_bad_value(self):
        """Test value to predefined structure pack is not an iterable of members."""
        exp_message = Baseline(
            """
             +--------+-------+-------+--------------------+
             | Offset | Value | Bytes | Format             |
             +--------+-------+-------+--------------------+
             |        |       |       | Custom (Structure) |
             +--------+-------+-------+--------------------+

             TypeError occurred during pack operation:

             cannot unpack non-iterable int object
             """
        )

        with pytest.raises(PackError) as trap:
            pack(0, Custom)

        assert wrap_message(trap.value) == exp_message
        assert isinstance(trap.value.__context__, TypeError)

    def test_pack_value_missing_member(self):
        """Test tuple value to pack is missing a required member."""
        with pytest.raises(PackError) as trap:
            pack((0,), Sample)

        exp_message = Baseline(
            """
             +--------+-------+-------+--------------------+
             | Offset | Value | Bytes | Format             |
             +--------+-------+-------+--------------------+
             |        |       |       | Sample (Structure) |
             +--------+-------+-------+--------------------+

             ValueError occurred during pack operation:

             not enough values to unpack (expected 2, got 1)
             """
        )

        assert wrap_message(trap.value) == exp_message
        assert isinstance(trap.value.__context__, ValueError)


class TestProtections:
    def test_byteorder(self):
        exp_message = Baseline(
            """
            byteorder must be either "big" or "little"
            """
        )

        with pytest.raises(ValueError) as trap:
            # pylint: disable=unused-variable
            class Struct(Structure, byteorder="wrong"):
                pass

        assert wrap_message(trap.value) == exp_message

    def test_fmt_arg_but_no_fmt_factory(self):
        exp_message = Baseline(
            """
            when 'fmt_arg' specified, 'fmt' must be a factory function
            """
        )

        arg = member(fmt=uint8)

        with pytest.raises(TypeError) as trap:
            member(fmt=uint8, fmt_arg=arg)

        assert wrap_message(trap.value) == exp_message

    def test_default_and_compute(self):
        exp_message = Baseline(
            """
            'default' may not be specified when 'compute=True'
            """
        )

        with pytest.raises(TypeError) as trap:
            member(fmt=uint8, compute=True, default=0)

        assert wrap_message(trap.value) == exp_message

    def test_setter_when_readonly(self):
        exp_message = Baseline(
            """
            'setter' not allowed on read-only structure member properties
            """
        )

        memb = member(fmt=uint8, readonly=True)

        with pytest.raises(TypeError) as trap:

            @memb.setter
            def memb(self, value):  # pylint: disable=unused-argument
                pass

        assert wrap_message(trap.value) == exp_message

    def test_deleter(self):
        exp_message = Baseline(
            """
            structure member properties do not support 'deleter'
            """
        )

        memb = member(fmt=uint8)

        with pytest.raises(TypeError) as trap:

            @memb.deleter
            def memb(self):  # pylint: disable=unused-argument
                pass

        assert wrap_message(trap.value) == exp_message


class TestGetterOverride:

    """Test explicit member getters retained."""

    class Struct(Structure):
        memb = member(fmt=uint8)

        @memb.getter
        def memb(self):
            return 99

    def test(self):
        struct = self.Struct(memb=0)
        assert struct.memb == 99


class Pet(IntEnum):
    CAT = 0
    DOG = 1


pet_x = EnumX(name="pet", enum=Pet, nbytes=1)


class StructWithEnum(Structure):
    pet: Pet = member(fmt=pet_x, default=Pet.CAT)


class TestMemberEnumDefault:

    """Test member with default being an enum member."""

    expected_dump = Baseline(
        """
        +--------+--------+---------+-------+----------------------------+
        | Offset | Access | Value   | Bytes | Format                     |
        +--------+--------+---------+-------+----------------------------+
        |        |        |         |       | StructWithEnum (Structure) |
        | 0      | pet    | Pet.CAT | 00    | pet                        |
        +--------+--------+---------+-------+----------------------------+
        """
    )

    def test_pack(self):
        struct = StructWithEnum()
        assert str(struct.dump) == self.expected_dump


class TestNestedFmt:

    """Test that format can be defined within class namespace."""

    expected_dump = Baseline(
        """
        +--------+-----------+-------+-------+--------------------+
        | Offset | Access    | Value | Bytes | Format             |
        +--------+-----------+-------+-------+--------------------+
        |        |           |       |       | Sample (Structure) |
        |        | sample1   |       |       | Nested (Structure) |
        | 0      |   nested1 | 1     | 01    | uint8              |
        | 1      |   nested2 | 2     | 02    | uint8              |
        +--------+-----------+-------+-------+--------------------+
        """
    )

    expected_bytes = bytes.fromhex("0102")

    class Sample(Structure):
        class Nested(Structure):
            nested1 = member(fmt=uint8)
            nested2 = member(fmt=uint8)

        sample1: int = member(fmt=Nested)

    def test_unpack_pack_roundtrip(self):
        sample = self.Sample.unpack(self.expected_bytes)
        assert sample.sample1.nested1 == 1
        assert sample.sample1.nested2 == 2
        assert sample.ipack() == self.expected_bytes

    def test_unpack_dump(self):
        sample, dump = self.Sample.unpack_and_dump(self.expected_bytes)
        assert sample.sample1.nested1 == 1
        assert sample.sample1.nested2 == 2
        assert str(dump) == self.expected_dump

    def test_pack_dump(self):
        sample = self.Sample.unpack(self.expected_bytes)
        assert str(sample.dump) == self.expected_dump



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Common typing hints."""

import sys
from typing import Any, Callable, Type, Union

from .data import Data
from .transform import Transform

if sys.version_info < (3, 8):
    ByteOrderHint = str
else:
    from typing import Literal

    ByteOrderHint = Union[Literal["little"], Literal["big"]]


Format = Union[Transform, Type[Data]]
FactoryFormat = Union[Callable[[Type[Any]], Format], Transform, Type[Data]]



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test SizedX transform API conformance."""

from plum.littleendian import uint8
from plum.sized import SizedX
from plum.str import StrX


ascii_greedy = StrX(name="ascii", encoding="ascii")

sized_string = SizedX(
    fmt=ascii_greedy,
    size_fmt=uint8,
    size_access="--nbytes--",
    name="sized_string",
)


class TestDefaults:

    xform = SizedX(fmt=ascii_greedy, size_fmt=uint8)

    def test_fmt(self):
        assert self.xform.fmt is ascii_greedy

    def test_hint(self):
        assert self.xform.__hint__ == "str"

    def test_name(self):
        assert self.xform.name == "sized: str"

    def test_offset(self):
        assert self.xform.offset == 0

    def test_ratio(self):
        assert self.xform.ratio == 1

    def test_size_access(self):
        assert self.xform.size_access == "--size--"

    def test_size_fmt(self):
        assert self.xform.size_fmt is uint8


class TestPositional:

    xform = SizedX(ascii_greedy, uint8, 2, 10, "size_access", "name")

    def test_fmt(self):
        assert self.xform.fmt is ascii_greedy

    def test_hint(self):
        assert self.xform.__hint__ == "str"

    def test_name(self):
        assert self.xform.name == "name"

    def test_offset(self):
        assert self.xform.offset == 10

    def test_ratio(self):
        assert self.xform.ratio == 2

    def test_size_access(self):
        assert self.xform.size_access == "size_access"

    def test_size_fmt(self):
        assert self.xform.size_fmt is uint8


class TestKeyword(TestPositional):

    xform = SizedX(
        fmt=ascii_greedy,
        size_fmt=uint8,
        ratio=2,
        offset=10,
        size_access="size_access",
        name="name",
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Bytes transform."""

from typing import List, Optional, Tuple, Union

from ._getbytes import getbytes
from .dump import Record
from .exceptions import InsufficientMemoryError
from .transform import Transform


class BytesX(Transform):

    """Bytes transform."""

    __np__: Tuple[Union[int, None], bytes]

    def __init__(
        self,
        nbytes: Optional[int] = None,
        pad: bytes = b"",
        name: Optional[str] = None,
    ) -> None:
        if name is None:
            if nbytes is None:
                descriptions = ["greedy"]
            else:
                descriptions = ["fixed"]

            if pad:
                descriptions.append("padded")

            name = f"bytes ({','.join(descriptions)})"

        super().__init__(name, hint="bytes")

        assert 0 <= len(pad) <= 1
        assert not (nbytes is None and pad), "must set nbytes when setting pad"

        self.__nbytes__ = nbytes
        self.__np__ = nbytes, pad

    @property
    def pad(self) -> bytes:
        """Pad byte."""
        return self.__np__[1]

    def __unpack__(
        self,
        buffer: bytes,
        offset: int,
        dump: Optional[Record] = None,
        nbytes: Optional[int] = None,
    ) -> Tuple[bytes, int]:
        # pylint: disable=arguments-differ
        if dump is not None:
            return self.__unpack_and_dump__(buffer, offset, dump, nbytes)

        nbytes_x, pad = self.__np__

        if nbytes is None:
            nbytes = nbytes_x

        if nbytes is None:
            value = bytes(buffer[offset:])
            offset += len(value)
        else:
            start = offset
            offset += nbytes
            value = bytes(buffer[start:offset])

            if len(value) < nbytes:
                raise InsufficientMemoryError("insufficient bytes")

            if pad:
                value = value.rstrip(pad)

        return value, offset

    def __unpack_and_dump__(
        self, buffer: bytes, offset: int, dump: Record, nbytes: Optional[int] = None
    ) -> Tuple[bytes, int]:
        # pylint: disable=arguments-differ
        nbytes_x, pad = self.__np__

        if nbytes is None:
            nbytes = nbytes_x

        try:
            value, offset = getbytes(buffer, offset, dump, nbytes)
        except InsufficientMemoryError:
            # getbytes() adds records in 16 byte chunks showing bytes it was able
            # to unpack, reassemble them as if operation was successful
            value = b"".join(subdump.memory for subdump in dump)
            del dump[:]
            for i in range(0, len(value), 16):
                chunk = value[i : i + 16]
                dump.add_record(
                    access=f"[{i}:{i + len(chunk)}]",
                    value=repr(chunk),
                    memory=chunk,
                )
            raise

        if nbytes is not None and pad:
            value = value.rstrip(pad)
            pad = pad * (nbytes - len(value))
        else:
            pad = b""

        dump.memory = b""
        if value:
            for i in range(0, len(value), 16):
                chunk = value[i : i + 16]
                dump.add_record(
                    access=f"[{i}:{i + len(chunk)}]",
                    value=repr(chunk),
                    memory=chunk,
                )
        else:
            dump.add_record(access="[0:0]", value=repr(value))

        for i in range(0, len(pad), 16):
            chunk = pad[i : i + 16]
            dump.add_record(
                access="--pad--",
                value=repr(chunk),
                memory=chunk,
            )

        return value, offset

    def __pack__(
        self,
        value: Union[bytes, bytearray],
        pieces: List[bytes],
        dump: Optional[Record] = None,
    ) -> None:
        if dump is not None:
            self.__pack_and_dump__(value, pieces, dump)

        else:
            nbytes, pad = self.__np__

            piece = bytes(value)

            if nbytes is not None:
                actual_nbytes = len(piece)

                if pad and actual_nbytes < nbytes:
                    piece += pad * (nbytes - actual_nbytes)

                elif actual_nbytes != nbytes:
                    raise ValueError(
                        f"expected length to be {nbytes} but instead found {actual_nbytes}"
                    )

            pieces.append(piece)

    def __pack_and_dump__(
        self,
        value: Union[bytes, bytearray],
        pieces: List[bytes],
        dump: Record,
    ) -> None:
        nbytes, pad = self.__np__

        piece = bytes(value)

        actual_nbytes = len(piece)

        if actual_nbytes:
            for i in range(0, actual_nbytes, 16):
                chunk = piece[i : i + 16]
                dump.add_record(
                    access=f"[{i}:{i + len(chunk)}]",
                    value=repr(chunk),
                    memory=chunk,
                )
        else:
            dump.add_record(access="[0:0]", value=repr(piece))

        if nbytes is not None and pad and actual_nbytes < nbytes:
            pad = pad * (nbytes - actual_nbytes)
            piece += pad

            for i in range(0, len(pad), 16):
                chunk = pad[i : i + 16]
                dump.add_record(
                    access="--pad--",
                    value=repr(chunk),
                    memory=chunk,
                )

        elif nbytes is not None and actual_nbytes != nbytes:
            raise ValueError(
                f"expected length to be {nbytes} but instead found {actual_nbytes}"
            )

        pieces.append(piece)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Structure type metaclass."""

from typing import Dict, Optional, Tuple

from .._code_injector import CodeInjector
from ..data import DataMeta
from .bitfield_member import BitFieldMember
from .code_maker import CodeMaker
from .member import Member


def get_offsets_and_nbytes(
    members: Dict[str, Member]
) -> Tuple[Optional[Tuple[int, ...]], Optional[int]]:
    """Compute member offsets and structure size."""
    member_sizes = tuple(member.nbytes for member in members.values())

    offset = 0
    offsets = []
    for size in member_sizes:
        offsets.append(offset)

        if size is None:
            # member variably sized, so new structure is as well
            nbytes = None
            member_offsets = None
            break

        offset += size
    else:
        member_offsets = tuple(offsets)
        nbytes = offset

    return member_offsets, nbytes


class StructureMeta(DataMeta):

    """Structure data store metaclass."""

    __byteorder__: str
    __members__: Tuple[Member, ...]

    def __new__(
        mcs,
        name,
        bases,
        namespace,
        byteorder: str = "little",
        fieldorder: str = "most_to_least",
    ):
        # pylint: disable=too-many-locals,too-many-branches

        namespace["__format_name__"] = (
            "Structure" if name == "Structure" else f"{name} (Structure)"
        )

        if byteorder not in {"big", "little"}:
            raise ValueError('byteorder must be either "big" or "little"')

        if fieldorder not in {"least_to_most", "most_to_least"}:
            raise ValueError(
                'fieldorder must be either "least_to_most" or "most_to_least"'
            )

        members_dict: Dict[str, Member] = {}
        for base_class in bases:
            if isinstance(base_class, StructureMeta):
                members_dict.update(
                    {m.name: m.copy(m.fget, m.fset) for m in base_class.__members__}
                )

        new_members: Dict[str, Member] = {
            k: m for k, m in namespace.items() if isinstance(m, Member)
        }

        if new_members:
            # namespace.update(new_members)
            members_dict.update(new_members)

            BitFieldMember.organize(members_dict, byteorder, fieldorder)

            code_injector = CodeInjector(namespace)

            index = 0
            for member_name, member in members_dict.items():
                index = member.add_name_index(member_name, index, code_injector)

            for member in members_dict.values():
                member.check()

            offsets, nbytes = get_offsets_and_nbytes(members_dict)

            members = tuple(members_dict.values())

            namespace["__ignore_flags__"] = tuple(member.ignore for member in members)
            namespace["__members__"] = members
            if "__nbytes__" not in namespace:  # pragma: no cover
                namespace["__nbytes__"] = nbytes
            namespace["__offsets__"] = offsets
            namespace["__byteorder__"] = byteorder
            if nbytes is not None:
                names = tuple(member.name for member in members)
                types = tuple(member.fmt for member in members)
                namespace["__names_types__"] = (names, types)

            code_maker = CodeMaker(name, members, new_members.values())
            lines = list(line.rstrip() for line in code_maker.iter_lines(namespace))

            code_injector.update_script(lines)
            code_injector.execute_lines(lines)
            namespace["__implementation__"] = "\n".join(lines)

        for member in members_dict.values():
            del member.temp_store

        cls = super().__new__(mcs, name, bases, namespace)

        # if boost and cls.__names_types__ != NO_MEMBERS_DEFINED:  # pragma: no cover
        #     for method_name in ['__pack__', '__unpack__']:
        #         if mcs._is_method_overridden(method_name, base_class, namespace):
        #             break
        #     else:
        #         nbytes = cls.__nbytes__
        #         _names, types = cls.__names_types__
        #
        #         # attach binary string containing plum-c accelerator "C" structure
        #         # (so structure memory de-allocated when class deleted)
        #         cls.__plum_boost_internals__ = boost.faststructure.add_c_acceleration(
        #             cls, -1 if nbytes is None else nbytes, len(types), types)
        #
        #         cls.__pack__ = boost.ipack_classmethod(cls.__pack__.__func__)
        #         cls.__unpack__ = boost.unpack_classmethod(cls.__unpack__.__func__)

        return cls

    @property
    def byteorder(cls) -> str:
        """Byte order ("little" or "big")."""
        return cls.__byteorder__



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Array to bytes and bytes to array transform."""

from typing import Any, Generator, List, Optional, Sequence, Tuple, Union

from ._typing import Format
from .data import DataMeta
from .dump import Record
from .int import IntX
from .transform import Transform

Dims = Sequence[Optional[int]]
DimsArg = Union[Dims, "ArrayX", IntX]

GREEDY_DIMS: Dims = (None,)


def iterate_dims(array: Sequence[Any], ndims: int) -> Generator[int, None, None]:
    """Determine dimensions from multi-dimensional array value."""
    for _ in range(ndims):
        yield len(array)
        array = array[0]


class ArrayX(Transform):

    """Array to bytes and bytes to array transform."""

    __ndims__: int
    __df__: Tuple[DimsArg, Format]

    def __init__(
        self,
        fmt: Format,
        dims: Optional[DimsArg] = None,
        name: Optional[str] = None,
    ) -> None:
        assert isinstance(fmt, (DataMeta, Transform))

        locked_dims: DimsArg

        num_dims = 1

        if dims is None:
            locked_dims = GREEDY_DIMS

        elif isinstance(dims, IntX):
            locked_dims = dims

        elif isinstance(dims, ArrayX):
            locked_dims = dims
            dims_dims, _fmt = dims.__df__

            if (
                not isinstance(dims_dims, tuple)
                or len(dims_dims) != 1
                or dims_dims[0] is None
            ):
                raise TypeError(
                    "invalid 'dims', when providing an array transform for `dims`, "
                    "it must have a single, fixed (non-greedy) dimension"
                )

            num_dims = dims_dims[0]

        else:
            locked_dims = tuple(None if d is None else int(d) for d in dims)
            assert all(True if d is None else d > 0 for d in locked_dims)
            num_dims = len(dims)

        nbytes: Optional[int] = fmt.__nbytes__

        if nbytes is not None:
            if isinstance(locked_dims, tuple):
                for dim in locked_dims:
                    if dim is None:
                        nbytes = None
                        break

                    nbytes *= dim

            else:
                nbytes = None

        hint = fmt.__hint__ or "Any"

        for _ in range(num_dims):
            hint = f"List[{hint}]"

        if name is None:
            name = hint

        super().__init__(name, hint)

        self.__df__ = locked_dims, fmt
        self.__ndims__ = num_dims
        self.__nbytes__ = nbytes

    @property
    def fmt(self) -> Format:
        """Array element format."""
        return self.__df__[1]

    @property
    def dims(self) -> DimsArg:
        """Array dimensions."""
        return self.__df__[0]

    def __unpack__(
        self,
        buffer: bytes,
        offset: int,
        dump: Optional[Record] = None,
        dims: Optional[Dims] = None,
    ) -> Tuple[List[Any], int]:
        # pylint: disable=arguments-differ,too-many-branches
        if dump is not None:
            return ArrayX.__unpack_and_dump__(self, buffer, offset, dump, dims)

        x_dims, fmt = self.__df__

        if dims is None:
            if isinstance(x_dims, tuple):
                dims = x_dims
                if None in dims[1:]:
                    raise TypeError(
                        "array unpack does not support greedy secondary dimensions"
                    )

            elif isinstance(x_dims, IntX):
                dim, offset = x_dims.__unpack__(buffer, offset, dump)
                dims = (dim,)

            elif isinstance(x_dims, ArrayX):
                # must be isinstance(x_dims, ArrayX):
                dims, offset = x_dims.__unpack__(buffer, offset, dump)

            else:  # pragma: no cover
                raise RuntimeError("unexpected implementation error")

        array: List[Any] = []
        append = array.append

        this_dim, *item_dims = dims

        if this_dim is None:
            end = len(buffer)
            while offset < end:
                item, offset = fmt.__unpack__(buffer, offset, dump)
                append(item)

        elif item_dims:
            for _ in range(this_dim):
                item, offset = ArrayX.__unpack__(self, buffer, offset, dump, item_dims)
                append(item)
        else:
            for _ in range(this_dim):
                item, offset = fmt.__unpack__(buffer, offset, dump)
                append(item)

        return array, offset

    def __unpack_and_dump__(
        self,
        buffer: bytes,
        offset: int,
        dump: Record,
        dims: Optional[Dims] = None,
    ) -> Tuple[List[Any], int]:
        # pylint: disable=arguments-differ,too-many-locals,too-many-branches
        x_dims, fmt = self.__df__

        if dims is None:
            if isinstance(x_dims, tuple):
                dims = x_dims
                if None in dims[1:]:
                    raise TypeError(
                        "array unpack does not support greedy secondary dimensions"
                    )

            elif isinstance(x_dims, IntX):
                dim, offset = x_dims.__unpack__(
                    buffer, offset, dump.add_record(access="len()", fmt=x_dims)
                )
                dims = (dim,)

            elif isinstance(x_dims, ArrayX):
                dims, offset = x_dims.__unpack__(
                    buffer, offset, dump.add_record(access="--dims--", fmt=x_dims)
                )

            else:  # pragma: no cover
                raise RuntimeError(f"invalid 'dims' {x_dims!r}")

        array: List[Any] = []
        append = array.append

        this_dim, *item_dims = dims

        if this_dim is None:
            end = len(buffer)
            i = 0
            fmt_name = fmt.name
            while offset < end:
                item, offset = fmt.__unpack__(
                    buffer, offset, dump.add_record(access=f"[{i}]", fmt=fmt_name)
                )
                append(item)
                i += 1

        elif item_dims:
            for i in range(this_dim):
                item, offset = ArrayX.__unpack_and_dump__(
                    self, buffer, offset, dump.add_record(access=f"[{i}]"), item_dims
                )
                append(item)
        else:
            fmt_name = fmt.name
            for i in range(this_dim):
                item, offset = fmt.__unpack__(
                    buffer, offset, dump.add_record(access=f"[{i}]", fmt=fmt_name)
                )
                append(item)

        if not array:
            dump.value = []

        return array, offset

    def __pack__(
        self,
        value: Sequence[Any],
        pieces: List[bytes],
        dump: Optional[Record] = None,
        dims: Optional[Dims] = None,
    ) -> None:
        # pylint: disable=arguments-differ,too-many-branches
        if dump is not None:
            ArrayX.__pack_and_dump__(self, value, pieces, dump, dims)

        else:
            x_dims, fmt = self.__df__

            if dims is None:
                if isinstance(x_dims, tuple):
                    this_dim, *item_dims = x_dims

                elif isinstance(x_dims, IntX):
                    this_dim, item_dims = len(value), []
                    x_dims.__pack__(this_dim, pieces, dump)

                elif isinstance(x_dims, ArrayX):
                    dims = tuple(iterate_dims(value, self.__ndims__))
                    this_dim, *item_dims = dims
                    x_dims.__pack__(dims, pieces, dump)

                else:  # pragma: no cover
                    raise RuntimeError(f"invalid 'dims' {x_dims!r}")

            else:
                this_dim, *item_dims = dims

            try:
                actual_length = len(value)
            except TypeError:
                raise TypeError("invalid array value, retrying with dump") from None

            if this_dim is not None and actual_length != this_dim:
                raise TypeError("invalid array value, retrying with dump")

            if item_dims:
                for item in value:
                    ArrayX.__pack__(self, item, pieces, dump, item_dims)
            else:
                for item in value:
                    fmt.__pack__(item, pieces, dump)

    def __pack_and_dump__(
        self,
        value: Sequence[Any],
        pieces: List[bytes],
        dump: Record,
        dims: Optional[Dims] = None,
    ) -> None:
        # pylint: disable=arguments-differ,too-many-branches
        x_dims, fmt = self.__df__

        if dims is None:
            if isinstance(x_dims, tuple):
                this_dim, *item_dims = x_dims

            elif isinstance(x_dims, IntX):
                this_dim, item_dims = len(value), []
                x_dims.__pack__(
                    this_dim, pieces, dump.add_record(access="len()", fmt=x_dims)
                )

            elif isinstance(x_dims, ArrayX):
                dims = tuple(iterate_dims(value, self.__ndims__))
                this_dim, *item_dims = dims
                x_dims.__pack__(
                    dims, pieces, dump.add_record(access="--dims--", fmt=x_dims)
                )

            else:  # pragma: no cover
                raise RuntimeError(f"invalid 'dims' {x_dims!r}")

        else:
            this_dim, *item_dims = dims

        try:
            actual_length = len(value)
        except TypeError:
            dump.value = value
            raise TypeError(
                f"invalid value, expected iterable of "
                f'{"any " if this_dim is None else ""}'
                f'length{"" if this_dim is None else " " + str(this_dim)}'
                f", got non-iterable"
            ) from None

        if not actual_length:
            dump.value = []

        if this_dim is None:
            this_dim = actual_length

        if item_dims:
            for i, item in zip(range(this_dim), value):
                ArrayX.__pack_and_dump__(
                    self, item, pieces, dump.add_record(access=f"[{i}]"), item_dims
                )
        else:
            fmt_name = fmt.name

            for i, item in zip(range(this_dim), value):
                fmt.__pack__(
                    item, pieces, dump.add_record(access=f"[{i}]", fmt=fmt_name)
                )

        if actual_length != this_dim:
            for i in range(actual_length, this_dim):
                dump.add_record(access=f"[{i}]", value="<missing>")

            for i, item in zip(range(this_dim, actual_length), value[this_dim:]):
                dump.add_record(
                    access=f"[{i}] <extra>", value=item, separate=(i == this_dim)
                )
            raise TypeError(
                f"invalid value, expected iterable of "
                f"{this_dim} length, got iterable of length {actual_length}"
            )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test BitFieldMember() member definition."""

# pylint: disable=too-many-lines

from enum import IntEnum
from baseline import Baseline

from plum.bigendian import uint8
from plum.conformance import Case, CaseData
from plum.structure import Structure, member, bitfield_member


class Pet(IntEnum):
    CAT = 0
    DOG = 1


class TestFirst(Case):

    """Bit fields are first members of structure."""

    class MyStruct(Structure, fieldorder="least_to_most"):

        m1: Pet = bitfield_member(typ=Pet, size=4)
        m2: int = bitfield_member(typ=int, size=4, default=1)
        m3: int = bitfield_member(typ=int, size=8, default=2)
        m4: int = member(fmt=uint8, default=0xFF)

    data = CaseData(
        fmt=MyStruct,
        bindata=b"\x21\x03\x04",
        nbytes=3,
        values=(
            MyStruct(m1=Pet.DOG, m2=2, m3=3, m4=4),
            [Pet.DOG, 2, 3, 4],
            (Pet.DOG, 2, 3, 4),
        ),
        dump=Baseline(
            """
            +---------+--------+---------+-------------------+----------------------+
            | Offset  | Access | Value   | Bytes             | Format               |
            +---------+--------+---------+-------------------+----------------------+
            |         |        |         |                   | MyStruct (Structure) |
            | 0       |        | 801     | 21 03             |                      |
            |  [0:4]  | m1     | Pet.DOG | ........ ....0001 | Pet                  |
            |  [4:8]  | m2     | 2       | ........ 0010.... | int                  |
            |  [8:16] | m3     | 3       | 00000011 ........ | int                  |
            | 2       | m4     | 4       | 04                | uint8                |
            +---------+--------+---------+-------------------+----------------------+
            """
        ),
        excess=Baseline(
            """
            +---------+--------+----------------+-------------------+----------------------+
            | Offset  | Access | Value          | Bytes             | Format               |
            +---------+--------+----------------+-------------------+----------------------+
            |         |        |                |                   | MyStruct (Structure) |
            | 0       |        | 801            | 21 03             |                      |
            |  [0:4]  | m1     | Pet.DOG        | ........ ....0001 | Pet                  |
            |  [4:8]  | m2     | 2              | ........ 0010.... | int                  |
            |  [8:16] | m3     | 3              | 00000011 ........ | int                  |
            | 2       | m4     | 4              | 04                | uint8                |
            +---------+--------+----------------+-------------------+----------------------+
            | 3       |        | <excess bytes> | 99                |                      |
            +---------+--------+----------------+-------------------+----------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +---------+--------+----------------------+-------------------+----------------------+
            | Offset  | Access | Value                | Bytes             | Format               |
            +---------+--------+----------------------+-------------------+----------------------+
            |         |        |                      |                   | MyStruct (Structure) |
            | 0       |        | 801                  | 21 03             |                      |
            |  [0:4]  | m1     | Pet.DOG              | ........ ....0001 | Pet                  |
            |  [4:8]  | m2     | 2                    | ........ 0010.... | int                  |
            |  [8:16] | m3     | 3                    | 00000011 ........ | int                  |
            |         | m4     | <insufficient bytes> |                   | uint8                |
            +---------+--------+----------------------+-------------------+----------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack uint8, 1 needed, only 0 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, m1: 'Pet', m2: Optional[int] = None, m3: Optional[int] = None, m4: Optional[int] = None) -> None:
                if m2 is None:
                    m2 = type(self).m2.default

                if m3 is None:
                    m3 = type(self).m3.default

                if m4 is None:
                    m4 = type(self).m4.default

                if not 0 <= m1 <= 15:
                   raise ValueError("'m1' out of range, 0 <= m1 <= 15")

                if not 0 <= m2 <= 15:
                   raise ValueError("'m2' out of range, 0 <= m2 <= 15")

                if not 0 <= m3 <= 255:
                   raise ValueError("'m3' out of range, 0 <= m3 <= 255")

                self[:] = (m1, m2, m3, m4)

            @m1.getter
            def m1(self) -> 'Pet':
                return self[0]

            @m1.setter
            def m1(self, value: 'Pet') -> None:
                self[0] = type(self).m1.typ(value)

            @m2.getter
            def m2(self) -> int:
                return self[1]

            @m2.setter
            def m2(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[1] = value

            @m3.getter
            def m3(self) -> int:
                return self[2]

            @m3.setter
            def m3(self, value: int) -> None:
                if not 0 <= value <= 255:
                   raise ValueError("out of range, 0 <= value <= 255")
                self[2] = value

            @m4.getter
            def m4(self) -> int:
                return self[3]

            @m4.setter
            def m4(self, value: int) -> None:
                self[3] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_m1, m_m2, m_m3, m_m4) = value

                if dump is None:
                    bitfields_0 = 0

                    bitfields_0 |= (bitfields_0 & -16) | (m_m1 & 0xf)

                    bitfields_0 |= (bitfields_0 & -241) | ((m_m2 & 0xf) << 4)

                    bitfields_0 |= (bitfields_0 & -65281) | ((m_m3 & 0xff) << 8)

                    pieces.append(bitfields_0.to_bytes(2, "little", signed=False))

                    cls.m4.fmt.__pack__(m_m4, pieces, dump)

                else:
                    bitfields_0_dump = dump.add_record()
                    bitfields_0 = 0

                    dump.add_record(access="m1", bits=(0, 4), value=m_m1, fmt=cls.m1.typ)
                    bitfields_0 |= (bitfields_0 & -16) | (m_m1 & 0xf)

                    dump.add_record(access="m2", bits=(4, 4), value=m_m2, fmt=int)
                    bitfields_0 |= (bitfields_0 & -241) | ((m_m2 & 0xf) << 4)

                    dump.add_record(access="m3", bits=(8, 8), value=m_m3, fmt=int)
                    bitfields_0 |= (bitfields_0 & -65281) | ((m_m3 & 0xff) << 8)

                    pieces.append(bitfields_0.to_bytes(2, "little", signed=False))
                    bitfields_0_dump.value = str(bitfields_0)
                    bitfields_0_dump.memory = pieces[-1]

                    m4_dump = dump.add_record(access="m4", fmt=cls.m4.fmt)
                    cls.m4.fmt.__pack__(m_m4, pieces, m4_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["MyStruct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    bitfields_0, offset = cls.m1.__unpack__(buffer, offset, dump, "little", nbytes=2, signed=False)

                    m_m1 = (bitfields_0 >> 0) & 0xf
                    m_m1 = cls.m1.typ(m_m1)

                    m_m2 = (bitfields_0 >> 4) & 0xf

                    m_m3 = (bitfields_0 >> 8) & 0xff

                    m_m4, offset = cls.m4.fmt.__unpack__(buffer, offset, dump)

                else:
                    bitfields_0_dump = dump.add_record()
                    bitfields_0, offset = cls.m1.__unpack__(buffer, offset, bitfields_0_dump, "little", nbytes=2, signed=False)

                    m_m1 = (bitfields_0 >> 0) & 0xf
                    m_m1 = cls.m1.typ(m_m1)
                    dump.add_record(access="m1", bits=(0, 4), value=m_m1, fmt=cls.m1.typ)

                    m_m2 = (bitfields_0 >> 4) & 0xf
                    dump.add_record(access="m2", bits=(4, 4), value=m_m2, fmt=int)

                    m_m3 = (bitfields_0 >> 8) & 0xff
                    dump.add_record(access="m3", bits=(8, 8), value=m_m3, fmt=int)

                    m4_dump = dump.add_record(access="m4", fmt=cls.m4.fmt)
                    m_m4, offset = cls.m4.fmt.__unpack__(buffer, offset, m4_dump)

                structure[:] = (m_m1, m_m2, m_m3, m_m4)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(m1={repr(self.m1).split(':', 1)[0].lstrip('<')}, m2={self.m2!r}, m3={self.m3!r}, m4={self.m4!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestLast(Case):
    """Bit fields are last members of structure."""

    class MyStruct(Structure, fieldorder="least_to_most"):

        m0: int = member(fmt=uint8, default=0xFF)
        m1: int = bitfield_member(typ=int, size=4)
        m2: int = bitfield_member(typ=int, size=4, default=1)
        m3: int = bitfield_member(typ=int, size=8, default=2)

    data = CaseData(
        fmt=MyStruct,
        bindata=b"\x00\x21\x03",
        nbytes=3,
        values=(MyStruct(m0=0, m1=1, m2=2, m3=3), [0, 1, 2, 3], (0, 1, 2, 3)),
        dump=Baseline(
            """
            +---------+--------+-------+-------------------+----------------------+
            | Offset  | Access | Value | Bytes             | Format               |
            +---------+--------+-------+-------------------+----------------------+
            |         |        |       |                   | MyStruct (Structure) |
            | 0       | m0     | 0     | 00                | uint8                |
            | 1       |        | 801   | 21 03             |                      |
            |  [0:4]  | m1     | 1     | ........ ....0001 | int                  |
            |  [4:8]  | m2     | 2     | ........ 0010.... | int                  |
            |  [8:16] | m3     | 3     | 00000011 ........ | int                  |
            +---------+--------+-------+-------------------+----------------------+
            """
        ),
        excess=Baseline(
            """
            +---------+--------+----------------+-------------------+----------------------+
            | Offset  | Access | Value          | Bytes             | Format               |
            +---------+--------+----------------+-------------------+----------------------+
            |         |        |                |                   | MyStruct (Structure) |
            | 0       | m0     | 0              | 00                | uint8                |
            | 1       |        | 801            | 21 03             |                      |
            |  [0:4]  | m1     | 1              | ........ ....0001 | int                  |
            |  [4:8]  | m2     | 2              | ........ 0010.... | int                  |
            |  [8:16] | m3     | 3              | 00000011 ........ | int                  |
            +---------+--------+----------------+-------------------+----------------------+
            | 3       |        | <excess bytes> | 99                |                      |
            +---------+--------+----------------+-------------------+----------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+--------+----------------------+-------+----------------------+
            | Offset | Access | Value                | Bytes | Format               |
            +--------+--------+----------------------+-------+----------------------+
            |        |        |                      |       | MyStruct (Structure) |
            | 0      | m0     | 0                    | 00    | uint8                |
            | 1      |        | <insufficient bytes> | 21    |                      |
            +--------+--------+----------------------+-------+----------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack, 2 needed, only 1 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, m0: Optional[int] = None, m1: int, m2: Optional[int] = None, m3: Optional[int] = None) -> None:
                if m0 is None:
                    m0 = type(self).m0.default

                if m2 is None:
                    m2 = type(self).m2.default

                if m3 is None:
                    m3 = type(self).m3.default

                if not 0 <= m1 <= 15:
                   raise ValueError("'m1' out of range, 0 <= m1 <= 15")

                if not 0 <= m2 <= 15:
                   raise ValueError("'m2' out of range, 0 <= m2 <= 15")

                if not 0 <= m3 <= 255:
                   raise ValueError("'m3' out of range, 0 <= m3 <= 255")

                self[:] = (m0, m1, m2, m3)

            @m0.getter
            def m0(self) -> int:
                return self[0]

            @m0.setter
            def m0(self, value: int) -> None:
                self[0] = value

            @m1.getter
            def m1(self) -> int:
                return self[1]

            @m1.setter
            def m1(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[1] = value

            @m2.getter
            def m2(self) -> int:
                return self[2]

            @m2.setter
            def m2(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[2] = value

            @m3.getter
            def m3(self) -> int:
                return self[3]

            @m3.setter
            def m3(self, value: int) -> None:
                if not 0 <= value <= 255:
                   raise ValueError("out of range, 0 <= value <= 255")
                self[3] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_m0, m_m1, m_m2, m_m3) = value

                if dump is None:
                    cls.m0.fmt.__pack__(m_m0, pieces, dump)

                    bitfields_0 = 0

                    bitfields_0 |= (bitfields_0 & -16) | (m_m1 & 0xf)

                    bitfields_0 |= (bitfields_0 & -241) | ((m_m2 & 0xf) << 4)

                    bitfields_0 |= (bitfields_0 & -65281) | ((m_m3 & 0xff) << 8)

                    pieces.append(bitfields_0.to_bytes(2, "little", signed=False))

                else:
                    m0_dump = dump.add_record(access="m0", fmt=cls.m0.fmt)
                    cls.m0.fmt.__pack__(m_m0, pieces, m0_dump)

                    bitfields_0_dump = dump.add_record()
                    bitfields_0 = 0

                    dump.add_record(access="m1", bits=(0, 4), value=m_m1, fmt=int)
                    bitfields_0 |= (bitfields_0 & -16) | (m_m1 & 0xf)

                    dump.add_record(access="m2", bits=(4, 4), value=m_m2, fmt=int)
                    bitfields_0 |= (bitfields_0 & -241) | ((m_m2 & 0xf) << 4)

                    dump.add_record(access="m3", bits=(8, 8), value=m_m3, fmt=int)
                    bitfields_0 |= (bitfields_0 & -65281) | ((m_m3 & 0xff) << 8)

                    pieces.append(bitfields_0.to_bytes(2, "little", signed=False))
                    bitfields_0_dump.value = str(bitfields_0)
                    bitfields_0_dump.memory = pieces[-1]

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["MyStruct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_m0, offset = cls.m0.fmt.__unpack__(buffer, offset, dump)

                    bitfields_0, offset = cls.m1.__unpack__(buffer, offset, dump, "little", nbytes=2, signed=False)

                    m_m1 = (bitfields_0 >> 0) & 0xf

                    m_m2 = (bitfields_0 >> 4) & 0xf

                    m_m3 = (bitfields_0 >> 8) & 0xff

                else:
                    m0_dump = dump.add_record(access="m0", fmt=cls.m0.fmt)
                    m_m0, offset = cls.m0.fmt.__unpack__(buffer, offset, m0_dump)

                    bitfields_0_dump = dump.add_record()
                    bitfields_0, offset = cls.m1.__unpack__(buffer, offset, bitfields_0_dump, "little", nbytes=2, signed=False)

                    m_m1 = (bitfields_0 >> 0) & 0xf
                    dump.add_record(access="m1", bits=(0, 4), value=m_m1, fmt=int)

                    m_m2 = (bitfields_0 >> 4) & 0xf
                    dump.add_record(access="m2", bits=(4, 4), value=m_m2, fmt=int)

                    m_m3 = (bitfields_0 >> 8) & 0xff
                    dump.add_record(access="m3", bits=(8, 8), value=m_m3, fmt=int)

                structure[:] = (m_m0, m_m1, m_m2, m_m3)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(m0={self.m0!r}, m1={self.m1!r}, m2={self.m2!r}, m3={self.m3!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestMultiple(Case):

    """Many bit field groups in structure."""

    class MyStruct(Structure, fieldorder="least_to_most"):

        m0: int = bitfield_member(typ=int, size=4)
        m1: int = bitfield_member(typ=int, size=8)

        m2: int = member(fmt=uint8)

        m3: int = bitfield_member(typ=int, size=4)
        m4: int = bitfield_member(typ=int, size=4)

        m5: int = member(fmt=uint8)

        m6: int = bitfield_member(typ=int, size=4)
        m7: int = bitfield_member(typ=int, size=8)

    data = CaseData(
        fmt=MyStruct,
        bindata=b"\x10\x00\x02\x43\x05\x76\x00",
        nbytes=7,
        values=(
            MyStruct(m0=0, m1=1, m2=2, m3=3, m4=4, m5=5, m6=6, m7=7),
            [0, 1, 2, 3, 4, 5, 6, 7],
            (0, 1, 2, 3, 4, 5, 6, 7),
        ),
        dump=Baseline(
            """
            +---------+--------+-------+-------------------+----------------------+
            | Offset  | Access | Value | Bytes             | Format               |
            +---------+--------+-------+-------------------+----------------------+
            |         |        |       |                   | MyStruct (Structure) |
            | 0       |        | 16    | 10 00             |                      |
            |  [0:4]  | m0     | 0     | ........ ....0000 | int                  |
            |  [4:12] | m1     | 1     | ....0000 0001.... | int                  |
            | 2       | m2     | 2     | 02                | uint8                |
            | 3       |        | 67    | 43                |                      |
            |  [0:4]  | m3     | 3     | ....0011          | int                  |
            |  [4:8]  | m4     | 4     | 0100....          | int                  |
            | 4       | m5     | 5     | 05                | uint8                |
            | 5       |        | 118   | 76 00             |                      |
            |  [0:4]  | m6     | 6     | ........ ....0110 | int                  |
            |  [4:12] | m7     | 7     | ....0000 0111.... | int                  |
            +---------+--------+-------+-------------------+----------------------+
            """
        ),
        excess=Baseline(
            """
            +---------+--------+----------------+-------------------+----------------------+
            | Offset  | Access | Value          | Bytes             | Format               |
            +---------+--------+----------------+-------------------+----------------------+
            |         |        |                |                   | MyStruct (Structure) |
            | 0       |        | 16             | 10 00             |                      |
            |  [0:4]  | m0     | 0              | ........ ....0000 | int                  |
            |  [4:12] | m1     | 1              | ....0000 0001.... | int                  |
            | 2       | m2     | 2              | 02                | uint8                |
            | 3       |        | 67             | 43                |                      |
            |  [0:4]  | m3     | 3              | ....0011          | int                  |
            |  [4:8]  | m4     | 4              | 0100....          | int                  |
            | 4       | m5     | 5              | 05                | uint8                |
            | 5       |        | 118            | 76 00             |                      |
            |  [0:4]  | m6     | 6              | ........ ....0110 | int                  |
            |  [4:12] | m7     | 7              | ....0000 0111.... | int                  |
            +---------+--------+----------------+-------------------+----------------------+
            | 7       |        | <excess bytes> | 99                |                      |
            +---------+--------+----------------+-------------------+----------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +---------+--------+----------------------+-------------------+----------------------+
            | Offset  | Access | Value                | Bytes             | Format               |
            +---------+--------+----------------------+-------------------+----------------------+
            |         |        |                      |                   | MyStruct (Structure) |
            | 0       |        | 16                   | 10 00             |                      |
            |  [0:4]  | m0     | 0                    | ........ ....0000 | int                  |
            |  [4:12] | m1     | 1                    | ....0000 0001.... | int                  |
            | 2       | m2     | 2                    | 02                | uint8                |
            | 3       |        | 67                   | 43                |                      |
            |  [0:4]  | m3     | 3                    | ....0011          | int                  |
            |  [4:8]  | m4     | 4                    | 0100....          | int                  |
            | 4       | m5     | 5                    | 05                | uint8                |
            | 5       |        | <insufficient bytes> | 76                |                      |
            +---------+--------+----------------------+-------------------+----------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack, 2 needed, only 1 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, m0: int, m1: int, m2: int, m3: int, m4: int, m5: int, m6: int, m7: int) -> None:
                if not 0 <= m0 <= 15:
                   raise ValueError("'m0' out of range, 0 <= m0 <= 15")

                if not 0 <= m1 <= 255:
                   raise ValueError("'m1' out of range, 0 <= m1 <= 255")

                if not 0 <= m3 <= 15:
                   raise ValueError("'m3' out of range, 0 <= m3 <= 15")

                if not 0 <= m4 <= 15:
                   raise ValueError("'m4' out of range, 0 <= m4 <= 15")

                if not 0 <= m6 <= 15:
                   raise ValueError("'m6' out of range, 0 <= m6 <= 15")

                if not 0 <= m7 <= 255:
                   raise ValueError("'m7' out of range, 0 <= m7 <= 255")

                self[:] = (m0, m1, m2, m3, m4, m5, m6, m7)

            @m0.getter
            def m0(self) -> int:
                return self[0]

            @m0.setter
            def m0(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[0] = value

            @m1.getter
            def m1(self) -> int:
                return self[1]

            @m1.setter
            def m1(self, value: int) -> None:
                if not 0 <= value <= 255:
                   raise ValueError("out of range, 0 <= value <= 255")
                self[1] = value

            @m2.getter
            def m2(self) -> int:
                return self[2]

            @m2.setter
            def m2(self, value: int) -> None:
                self[2] = value

            @m3.getter
            def m3(self) -> int:
                return self[3]

            @m3.setter
            def m3(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[3] = value

            @m4.getter
            def m4(self) -> int:
                return self[4]

            @m4.setter
            def m4(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[4] = value

            @m5.getter
            def m5(self) -> int:
                return self[5]

            @m5.setter
            def m5(self, value: int) -> None:
                self[5] = value

            @m6.getter
            def m6(self) -> int:
                return self[6]

            @m6.setter
            def m6(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[6] = value

            @m7.getter
            def m7(self) -> int:
                return self[7]

            @m7.setter
            def m7(self, value: int) -> None:
                if not 0 <= value <= 255:
                   raise ValueError("out of range, 0 <= value <= 255")
                self[7] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_m0, m_m1, m_m2, m_m3, m_m4, m_m5, m_m6, m_m7) = value

                if dump is None:
                    bitfields_0 = 0

                    bitfields_0 |= (bitfields_0 & -16) | (m_m0 & 0xf)

                    bitfields_0 |= (bitfields_0 & -4081) | ((m_m1 & 0xff) << 4)

                    pieces.append(bitfields_0.to_bytes(2, "little", signed=False))

                    cls.m2.fmt.__pack__(m_m2, pieces, dump)

                    bitfields_1 = 0

                    bitfields_1 |= (bitfields_1 & -16) | (m_m3 & 0xf)

                    bitfields_1 |= (bitfields_1 & -241) | ((m_m4 & 0xf) << 4)

                    pieces.append(bitfields_1.to_bytes(1, "little", signed=False))

                    cls.m5.fmt.__pack__(m_m5, pieces, dump)

                    bitfields_2 = 0

                    bitfields_2 |= (bitfields_2 & -16) | (m_m6 & 0xf)

                    bitfields_2 |= (bitfields_2 & -4081) | ((m_m7 & 0xff) << 4)

                    pieces.append(bitfields_2.to_bytes(2, "little", signed=False))

                else:
                    bitfields_0_dump = dump.add_record()
                    bitfields_0 = 0

                    dump.add_record(access="m0", bits=(0, 4), value=m_m0, fmt=int)
                    bitfields_0 |= (bitfields_0 & -16) | (m_m0 & 0xf)

                    dump.add_record(access="m1", bits=(4, 8), value=m_m1, fmt=int)
                    bitfields_0 |= (bitfields_0 & -4081) | ((m_m1 & 0xff) << 4)

                    pieces.append(bitfields_0.to_bytes(2, "little", signed=False))
                    bitfields_0_dump.value = str(bitfields_0)
                    bitfields_0_dump.memory = pieces[-1]

                    m2_dump = dump.add_record(access="m2", fmt=cls.m2.fmt)
                    cls.m2.fmt.__pack__(m_m2, pieces, m2_dump)

                    bitfields_1_dump = dump.add_record()
                    bitfields_1 = 0

                    dump.add_record(access="m3", bits=(0, 4), value=m_m3, fmt=int)
                    bitfields_1 |= (bitfields_1 & -16) | (m_m3 & 0xf)

                    dump.add_record(access="m4", bits=(4, 4), value=m_m4, fmt=int)
                    bitfields_1 |= (bitfields_1 & -241) | ((m_m4 & 0xf) << 4)

                    pieces.append(bitfields_1.to_bytes(1, "little", signed=False))
                    bitfields_1_dump.value = str(bitfields_1)
                    bitfields_1_dump.memory = pieces[-1]

                    m5_dump = dump.add_record(access="m5", fmt=cls.m5.fmt)
                    cls.m5.fmt.__pack__(m_m5, pieces, m5_dump)

                    bitfields_2_dump = dump.add_record()
                    bitfields_2 = 0

                    dump.add_record(access="m6", bits=(0, 4), value=m_m6, fmt=int)
                    bitfields_2 |= (bitfields_2 & -16) | (m_m6 & 0xf)

                    dump.add_record(access="m7", bits=(4, 8), value=m_m7, fmt=int)
                    bitfields_2 |= (bitfields_2 & -4081) | ((m_m7 & 0xff) << 4)

                    pieces.append(bitfields_2.to_bytes(2, "little", signed=False))
                    bitfields_2_dump.value = str(bitfields_2)
                    bitfields_2_dump.memory = pieces[-1]

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["MyStruct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    bitfields_0, offset = cls.m0.__unpack__(buffer, offset, dump, "little", nbytes=2, signed=False)

                    m_m0 = (bitfields_0 >> 0) & 0xf

                    m_m1 = (bitfields_0 >> 4) & 0xff

                    m_m2, offset = cls.m2.fmt.__unpack__(buffer, offset, dump)

                    bitfields_1, offset = cls.m3.__unpack__(buffer, offset, dump, "little", nbytes=1, signed=False)

                    m_m3 = (bitfields_1 >> 0) & 0xf

                    m_m4 = (bitfields_1 >> 4) & 0xf

                    m_m5, offset = cls.m5.fmt.__unpack__(buffer, offset, dump)

                    bitfields_2, offset = cls.m6.__unpack__(buffer, offset, dump, "little", nbytes=2, signed=False)

                    m_m6 = (bitfields_2 >> 0) & 0xf

                    m_m7 = (bitfields_2 >> 4) & 0xff

                else:
                    bitfields_0_dump = dump.add_record()
                    bitfields_0, offset = cls.m0.__unpack__(buffer, offset, bitfields_0_dump, "little", nbytes=2, signed=False)

                    m_m0 = (bitfields_0 >> 0) & 0xf
                    dump.add_record(access="m0", bits=(0, 4), value=m_m0, fmt=int)

                    m_m1 = (bitfields_0 >> 4) & 0xff
                    dump.add_record(access="m1", bits=(4, 8), value=m_m1, fmt=int)

                    m2_dump = dump.add_record(access="m2", fmt=cls.m2.fmt)
                    m_m2, offset = cls.m2.fmt.__unpack__(buffer, offset, m2_dump)

                    bitfields_1_dump = dump.add_record()
                    bitfields_1, offset = cls.m3.__unpack__(buffer, offset, bitfields_1_dump, "little", nbytes=1, signed=False)

                    m_m3 = (bitfields_1 >> 0) & 0xf
                    dump.add_record(access="m3", bits=(0, 4), value=m_m3, fmt=int)

                    m_m4 = (bitfields_1 >> 4) & 0xf
                    dump.add_record(access="m4", bits=(4, 4), value=m_m4, fmt=int)

                    m5_dump = dump.add_record(access="m5", fmt=cls.m5.fmt)
                    m_m5, offset = cls.m5.fmt.__unpack__(buffer, offset, m5_dump)

                    bitfields_2_dump = dump.add_record()
                    bitfields_2, offset = cls.m6.__unpack__(buffer, offset, bitfields_2_dump, "little", nbytes=2, signed=False)

                    m_m6 = (bitfields_2 >> 0) & 0xf
                    dump.add_record(access="m6", bits=(0, 4), value=m_m6, fmt=int)

                    m_m7 = (bitfields_2 >> 4) & 0xff
                    dump.add_record(access="m7", bits=(4, 8), value=m_m7, fmt=int)

                structure[:] = (m_m0, m_m1, m_m2, m_m3, m_m4, m_m5, m_m6, m_m7)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(m0={self.m0!r}, m1={self.m1!r}, m2={self.m2!r}, m3={self.m3!r}, m4={self.m4!r}, m5={self.m5!r}, m6={self.m6!r}, m7={self.m7!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestArtificialSplit(Case):

    """Many bit fields get grouped separately when using nbytes."""

    class MyStruct(Structure, fieldorder="least_to_most"):
        m1: int = bitfield_member(typ=int, size=4)
        m2: int = bitfield_member(typ=int, size=2)

        m3: int = bitfield_member(typ=int, size=4, nbytes=2)
        m4: int = bitfield_member(typ=int, size=4)

    data = CaseData(
        fmt=MyStruct,
        bindata=b"\x21\x43\x00",
        nbytes=3,
        values=(MyStruct(m1=1, m2=2, m3=3, m4=4), [1, 2, 3, 4], (1, 2, 3, 4)),
        dump=Baseline(
            """
            +--------+--------+-------+-------------------+----------------------+
            | Offset | Access | Value | Bytes             | Format               |
            +--------+--------+-------+-------------------+----------------------+
            |        |        |       |                   | MyStruct (Structure) |
            | 0      |        | 33    | 21                |                      |
            |  [0:4] | m1     | 1     | ....0001          | int                  |
            |  [4:6] | m2     | 2     | ..10....          | int                  |
            | 1      |        | 67    | 43 00             |                      |
            |  [0:4] | m3     | 3     | ........ ....0011 | int                  |
            |  [4:8] | m4     | 4     | ........ 0100.... | int                  |
            +--------+--------+-------+-------------------+----------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+--------+----------------+-------------------+----------------------+
            | Offset | Access | Value          | Bytes             | Format               |
            +--------+--------+----------------+-------------------+----------------------+
            |        |        |                |                   | MyStruct (Structure) |
            | 0      |        | 33             | 21                |                      |
            |  [0:4] | m1     | 1              | ....0001          | int                  |
            |  [4:6] | m2     | 2              | ..10....          | int                  |
            | 1      |        | 67             | 43 00             |                      |
            |  [0:4] | m3     | 3              | ........ ....0011 | int                  |
            |  [4:8] | m4     | 4              | ........ 0100.... | int                  |
            +--------+--------+----------------+-------------------+----------------------+
            | 3      |        | <excess bytes> | 99                |                      |
            +--------+--------+----------------+-------------------+----------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+--------+----------------------+----------+----------------------+
            | Offset | Access | Value                | Bytes    | Format               |
            +--------+--------+----------------------+----------+----------------------+
            |        |        |                      |          | MyStruct (Structure) |
            | 0      |        | 33                   | 21       |                      |
            |  [0:4] | m1     | 1                    | ....0001 | int                  |
            |  [4:6] | m2     | 2                    | ..10.... | int                  |
            | 1      |        | <insufficient bytes> | 43       |                      |
            +--------+--------+----------------------+----------+----------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack, 2 needed, only 1 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, m1: int, m2: int, m3: int, m4: int) -> None:
                if not 0 <= m1 <= 15:
                   raise ValueError("'m1' out of range, 0 <= m1 <= 15")

                if not 0 <= m2 <= 3:
                   raise ValueError("'m2' out of range, 0 <= m2 <= 3")

                if not 0 <= m3 <= 15:
                   raise ValueError("'m3' out of range, 0 <= m3 <= 15")

                if not 0 <= m4 <= 15:
                   raise ValueError("'m4' out of range, 0 <= m4 <= 15")

                self[:] = (m1, m2, m3, m4)

            @m1.getter
            def m1(self) -> int:
                return self[0]

            @m1.setter
            def m1(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[0] = value

            @m2.getter
            def m2(self) -> int:
                return self[1]

            @m2.setter
            def m2(self, value: int) -> None:
                if not 0 <= value <= 3:
                   raise ValueError("out of range, 0 <= value <= 3")
                self[1] = value

            @m3.getter
            def m3(self) -> int:
                return self[2]

            @m3.setter
            def m3(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[2] = value

            @m4.getter
            def m4(self) -> int:
                return self[3]

            @m4.setter
            def m4(self, value: int) -> None:
                if not 0 <= value <= 15:
                   raise ValueError("out of range, 0 <= value <= 15")
                self[3] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_m1, m_m2, m_m3, m_m4) = value

                if dump is None:
                    bitfields_0 = 0

                    bitfields_0 |= (bitfields_0 & -16) | (m_m1 & 0xf)

                    bitfields_0 |= (bitfields_0 & -49) | ((m_m2 & 0x3) << 4)

                    pieces.append(bitfields_0.to_bytes(1, "little", signed=False))

                    bitfields_1 = 0

                    bitfields_1 |= (bitfields_1 & -16) | (m_m3 & 0xf)

                    bitfields_1 |= (bitfields_1 & -241) | ((m_m4 & 0xf) << 4)

                    pieces.append(bitfields_1.to_bytes(2, "little", signed=False))

                else:
                    bitfields_0_dump = dump.add_record()
                    bitfields_0 = 0

                    dump.add_record(access="m1", bits=(0, 4), value=m_m1, fmt=int)
                    bitfields_0 |= (bitfields_0 & -16) | (m_m1 & 0xf)

                    dump.add_record(access="m2", bits=(4, 2), value=m_m2, fmt=int)
                    bitfields_0 |= (bitfields_0 & -49) | ((m_m2 & 0x3) << 4)

                    pieces.append(bitfields_0.to_bytes(1, "little", signed=False))
                    bitfields_0_dump.value = str(bitfields_0)
                    bitfields_0_dump.memory = pieces[-1]

                    bitfields_1_dump = dump.add_record()
                    bitfields_1 = 0

                    dump.add_record(access="m3", bits=(0, 4), value=m_m3, fmt=int)
                    bitfields_1 |= (bitfields_1 & -16) | (m_m3 & 0xf)

                    dump.add_record(access="m4", bits=(4, 4), value=m_m4, fmt=int)
                    bitfields_1 |= (bitfields_1 & -241) | ((m_m4 & 0xf) << 4)

                    pieces.append(bitfields_1.to_bytes(2, "little", signed=False))
                    bitfields_1_dump.value = str(bitfields_1)
                    bitfields_1_dump.memory = pieces[-1]

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["MyStruct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    bitfields_0, offset = cls.m1.__unpack__(buffer, offset, dump, "little", nbytes=1, signed=False)

                    m_m1 = (bitfields_0 >> 0) & 0xf

                    m_m2 = (bitfields_0 >> 4) & 0x3

                    bitfields_1, offset = cls.m3.__unpack__(buffer, offset, dump, "little", nbytes=2, signed=False)

                    m_m3 = (bitfields_1 >> 0) & 0xf

                    m_m4 = (bitfields_1 >> 4) & 0xf

                else:
                    bitfields_0_dump = dump.add_record()
                    bitfields_0, offset = cls.m1.__unpack__(buffer, offset, bitfields_0_dump, "little", nbytes=1, signed=False)

                    m_m1 = (bitfields_0 >> 0) & 0xf
                    dump.add_record(access="m1", bits=(0, 4), value=m_m1, fmt=int)

                    m_m2 = (bitfields_0 >> 4) & 0x3
                    dump.add_record(access="m2", bits=(4, 2), value=m_m2, fmt=int)

                    bitfields_1_dump = dump.add_record()
                    bitfields_1, offset = cls.m3.__unpack__(buffer, offset, bitfields_1_dump, "little", nbytes=2, signed=False)

                    m_m3 = (bitfields_1 >> 0) & 0xf
                    dump.add_record(access="m3", bits=(0, 4), value=m_m3, fmt=int)

                    m_m4 = (bitfields_1 >> 4) & 0xf
                    dump.add_record(access="m4", bits=(4, 4), value=m_m4, fmt=int)

                structure[:] = (m_m1, m_m2, m_m3, m_m4)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(m1={self.m1!r}, m2={self.m2!r}, m3={self.m3!r}, m4={self.m4!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestTypDefault:

    """Test typ parameter default behavior."""

    def test_defaults_to_int(self):
        """Test defaults to int when no annotation."""

        class Struct(Structure):

            """Sample Structure."""

            m1 = bitfield_member(size=4)

        assert Struct.m1.typ is int



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test construction of bit field classes."""

# pylint: disable=invalid-name

import pytest
from baseline import Baseline

from plum.bitfields import BitFields, bitfield
from plum.conformance import wrap_message


class TestMethodOverrides:

    """Test __init__ and property getters left alone if supplied in class namespace."""

    def test_init(self):
        class MyBits(BitFields):
            f0 = bitfield(size=1)

            def __init__(self, custom):
                super().__init__()
                MyBits.custom = custom

        MyBits(custom=True)
        assert MyBits.custom is True

    def test_getter(self):
        class MyBits(BitFields):
            f0 = bitfield(size=1)

            @f0.getter
            def f0(self):
                return "custom"

        assert MyBits.from_int(0).f0 == "custom"


class TestExceptions:

    """Test validation of class."""

    @staticmethod
    def test_bitfield_bad_type():
        """Test annotation contains bad type."""

        expected_message = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:

            class Sample(BitFields):  # pylint: disable=unused-variable
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=8)
                f2: int = bitfield(lsb=8, size=8, typ=str)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_bitfield_overlap():
        """Test bit fields overlap."""

        expected_message = Baseline(
            """
            bit field 'f2' overlaps with 'f1'
            """
        )

        with pytest.raises(TypeError) as trap:

            class Sample(BitFields):  # pylint: disable=unused-variable
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=8)
                f2: int = bitfield(lsb=7, size=8)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_nbytes_too_small():
        """Test nbytes argument insufficient size."""

        expected_message = Baseline(
            """
            nbytes must be at least 2 for bitfields specified
            """
        )

        with pytest.raises(TypeError) as trap:

            class Sample(BitFields, nbytes=1):  # pylint: disable=unused-variable
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=8)
                f2: int = bitfield(lsb=8, size=8)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_byteorder_invalid():
        """Test byteorder argument wrong value."""

        expected_message = Baseline(
            """
            byteorder must be either "big" or "little"
            """
        )

        with pytest.raises(ValueError) as trap:
            # pylint: disable=unused-variable
            class Sample(BitFields, byteorder="native"):
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=8)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_default_too_big():
        """Test default argument exceeds max."""

        expected_message = Baseline(
            """
            default must be: 0 <= number <= 0xff
            """
        )

        with pytest.raises(ValueError) as trap:
            # pylint: disable=unused-variable
            class Sample(BitFields, nbytes=1, default=0x100):
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=8)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_default_negative():
        """Test default argument less than zero."""

        expected_message = Baseline(
            """
            default must be: 0 <= number <= 0xff
            """
        )

        with pytest.raises(ValueError) as trap:
            # pylint: disable=unused-variable
            class Sample(BitFields, nbytes=1, default=-1):
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=1, default=1)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_ignore_too_big():
        """Test ignore argument exceeds max."""

        expected_message = Baseline(
            """
            ignore must be: 0 <= number <= 0xff
            """
        )

        with pytest.raises(ValueError) as trap:
            # pylint: disable=unused-variable
            class Sample(BitFields, nbytes=1, ignore=0x100):
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=8)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_ignore_negative():
        """Test ignore argument less than zero."""

        expected_message = Baseline(
            """
            ignore must be: 0 <= number <= 0xff
            """
        )

        with pytest.raises(ValueError) as trap:

            class Sample(
                BitFields, nbytes=1, ignore=-1
            ):  # pylint: disable=unused-variable
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=1)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_fieldorder_invalid():
        """Test fieldorder argument wrong value."""

        expected_message = Baseline(
            """
            fieldorder must be either "least_to_most" or "most_to_least"
            """
        )

        with pytest.raises(ValueError) as trap:
            # pylint: disable=unused-variable
            class Sample(BitFields, fieldorder="not_right"):
                """Sample bit field type."""

                f1: int = bitfield(lsb=0, size=8)

        assert wrap_message(trap.value) == expected_message


class TestAutoPositionLeast:

    """Test position automatically assigned when not provided (least_to_most)."""

    class AutoAll(BitFields, nbytes=1, fieldorder="least_to_most"):
        """Sample BitFields subclass with automatic positions for every field."""

        f1: int = bitfield(size=2)
        f2: int = bitfield(size=5)
        f3: int = bitfield(size=1)

    class AutoMix(BitFields, nbytes=1, fieldorder="least_to_most"):
        """Sample BitFields subclass with automatic positions for some fields."""

        f1: int = bitfield(lsb=0, size=2)
        f2: int = bitfield(size=5)
        f3: int = bitfield(lsb=7, size=1)

    def test_all_missing(self):
        """Test fills in all lsbs."""
        assert self.AutoAll.f1.lsb == 0
        assert self.AutoAll.f2.lsb == 2
        assert self.AutoAll.f3.lsb == 7
        assert repr(self.AutoAll.from_int(0x8B)) == "AutoAll(f1=3, f2=2, f3=1)"

    def test_mix(self):
        """Test fills in missing lsbs (some present)."""
        assert self.AutoMix.f1.lsb == 0
        assert self.AutoMix.f2.lsb == 2
        assert self.AutoMix.f3.lsb == 7
        assert repr(self.AutoMix.from_int(0x8B)) == "AutoMix(f1=3, f2=2, f3=1)"


class TestAutoPositionMost:

    """Test position automatically assigned when not provided (most_to_least)."""

    class AutoAll(BitFields):
        """Sample BitFields subclass with automatic positions for every field."""

        f3: int = bitfield(size=1)
        f2: int = bitfield(size=5)
        f1: int = bitfield(size=2)

    class AutoMix(BitFields):
        """Sample BitFields subclass with automatic positions for some fields."""

        f3: int = bitfield(lsb=7, size=1)
        f2: int = bitfield(size=5)
        f1: int = bitfield(lsb=0, size=2)

    def test_all_missing(self):
        """Test fills in all lsbs."""
        assert self.AutoAll.f1.lsb == 0
        assert self.AutoAll.f2.lsb == 2
        assert self.AutoAll.f3.lsb == 7
        assert repr(self.AutoAll.from_int(0x8B)) == "AutoAll(f3=1, f2=2, f1=3)"

    def test_mix(self):
        """Test fills in missing lsbs (some present)."""
        assert self.AutoAll.f1.lsb == 0
        assert self.AutoAll.f2.lsb == 2
        assert self.AutoAll.f3.lsb == 7
        assert repr(self.AutoMix.from_int(0x8B)) == "AutoMix(f3=1, f2=2, f1=3)"


class TestMisc:
    def test_no_typ(self):
        """Test bit field definition has no typ specified or type annotation."""

        class Sample(BitFields, nbytes=1):  # pylint: disable=unused-variable
            """Sample bit fields."""

            xyz = bitfield(lsb=0, size=1)

        assert Sample.xyz.typ is int

    def test_autosize(self):
        """Test nbytes property automatically computed properly."""

        class Sample(BitFields):
            """Sample bit fields."""

            abc = bitfield(lsb=22, size=1)
            xyz = bitfield(lsb=0, size=1)

        assert Sample.nbytes == 3  # pylint: disable=comparison-with-callable



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test transform base class."""

# anomaly in pylint: disable=no-member

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import ExcessMemoryError, UnpackError
from plum.littleendian import uint16


class TestExcessBytes:

    """Test unpack dump properly formats larger quantity of extra bytes."""

    def test_unpack_error(self):
        """Exercise formatting all of the excessive bytes."""
        expected_message = Baseline(
            """
            +--------+----------------+-------------------------------------------------+--------+
            | Offset | Value          | Bytes                                           | Format |
            +--------+----------------+-------------------------------------------------+--------+
            |  0     | 39321          | 99 99                                           | uint16 |
            +--------+----------------+-------------------------------------------------+--------+
            |  2     | <excess bytes> | 99 99 99 99 99 99 99 99 99 99 99 99 99 99 99 99 |        |
            | 18     |                | 99 99 99 99                                     |        |
            +--------+----------------+-------------------------------------------------+--------+

            ExcessMemoryError occurred during unpack operation:

            20 unconsumed bytes
            """
        )
        with pytest.raises(UnpackError) as trap:
            uint16.unpack(b"\x99" * 22)

        assert wrap_message(trap.value) == expected_message
        assert isinstance(trap.value.__context__, ExcessMemoryError)


class TestStrRepr:
    def test_str_matches_repr(self):
        assert str(uint16) == repr(uint16)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2023 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test integer flag dump without names."""

from baseline import Baseline

from plum.enum import EnumX
from sample_flag import Register, RegisterMixin

register = EnumX(Register, 1, "big", signed=False, strict=False)
register_mixin = EnumX(RegisterMixin, 1, "big", signed=False, strict=False)


class TestIntEnum:

    """Test FlagX default name with IntFlag."""

    expected_dump = Baseline(
        """
        +--------+-------------+-------+--------------------+
        | Offset | Value       | Bytes | Format             |
        +--------+-------------+-------+--------------------+
        | 0      | Register.SP | 01    | Register (IntFlag) |
        +--------+-------------+-------+--------------------+
        """
    )

    def test_pack_and_dump(self):
        _, dump = register.pack_and_dump(1)
        assert str(dump) == self.expected_dump


class TestIntMixinEnum:

    """Test FlagX default name with int, Flag mixin."""

    expected_dump = Baseline(
        """
        +--------+------------------+-------+---------------------------+
        | Offset | Value            | Bytes | Format                    |
        +--------+------------------+-------+---------------------------+
        | 0      | RegisterMixin.SP | 01    | RegisterMixin (int, Flag) |
        +--------+------------------+-------+---------------------------+
        """
    )

    def test_pack_and_dump(self):
        _, dump = register_mixin.pack_and_dump(1)
        assert str(dump) == self.expected_dump



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""plum-py package builder/installer."""

from setuptools import setup, find_packages

LONG_DESCRIPTION = """
#########################
[plum] Pack/Unpack Memory
#########################

.. image:: https://readthedocs.org/projects/plum-py/badge/?version=latest
    :target: https://plum-py.readthedocs.io/en/latest/?badge=latest
    :alt: Documentation Status

The plum-py Python package provides classes and utility functions to
transform byte sequences into Python objects and back. While similar in
purpose to Python's standard library ``struct`` module, this package
provides a larger set of format specifiers and is extensible, allowing
you to easily create complex ones of your own.
"""

setup(
    name="plum-py",
    version="0.8.7",
    description="Pack/Unpack Memory.",
    long_description=LONG_DESCRIPTION,
    url="https://plum-py.readthedocs.io/en/latest/index.html",
    download_url="https://plum-py.readthedocs.io/en/latest/index.html",
    author="Dan Gass",
    author_email="dan.gass@gmail.com",
    maintainer="Dan Gass",
    maintainer_email="dan.gass@gmail.com",
    license="MIT License; http://opensource.org/licenses/MIT",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
    package_dir={"": "src"},
    packages=find_packages("src"),
    package_data={
        "plum": ["py.typed"],
    },
)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test miscellaneous features of AttrDict and AttrDictX transform."""

import pytest

from baseline import Baseline

from plum.attrdict import AttrDict, AttrDictX
from plum.array import ArrayX
from plum.conformance import wrap_message
from plum.exceptions import SizeError
from plum.littleendian import uint8, uint16


class TestAttrDictX:

    """Test AttrDictX features."""

    def test_default_name(self):
        xform = AttrDictX(dict(m1=uint8, m2=uint16))
        assert xform.name == "AttrDict"

    def test_explicit_name(self):
        xform = AttrDictX(dict(m1=uint8, m2=uint16), name="name")
        assert xform.name == "name"

    def test_default_hint(self):
        xform = AttrDictX(dict(m1=uint8, m2=uint16))
        assert xform.__hint__ == "AttrDict"

    def test_size_error(self):
        """Verify no size when a member has no size."""
        expected_message = Baseline(
            """
            <transform 'AttrDict'> sizes vary
            """
        )

        xform = AttrDictX({"m1": ArrayX(name="list", fmt=uint8), "m2": uint8})

        with pytest.raises(SizeError) as trap:
            xform.nbytes  # pylint: disable=pointless-statement

        assert wrap_message(trap.value) == expected_message


class TestAttrDict:

    """Test AttrDict properties."""

    def test_attribute_access(self):
        dct = AttrDict(m1=1)
        assert dct["m1"] == 1
        assert dct.m1 == 1  # pylint: disable=no-member

    def test_repr(self):
        expected_repr = Baseline(
            """
            AttrDict(m1=1, m2=2)
            """
        )

        assert repr(AttrDict(m1=1, m2=2)) == expected_repr


class TestEmptyValueInDump:

    xform = AttrDictX({})

    expected_dump = Baseline(
        """
        +--------+-------+-------+----------+
        | Offset | Value | Bytes | Format   |
        +--------+-------+-------+----------+
        |        | {}    |       | AttrDict |
        +--------+-------+-------+----------+
        """
    )

    def test_pack(self):
        buffer, dump = self.xform.pack_and_dump({})
        assert buffer == b""
        assert str(dump) == self.expected_dump

    def test_unpack(self):
        value, dump = self.xform.unpack_and_dump(b"")
        assert value == {}
        assert str(dump) == self.expected_dump



"""Test decimal transform exceptions."""

# pylint: disable=unidiomatic-typecheck
import pytest

from baseline import Baseline

from plum.decimal import DecimalX
from plum.conformance import wrap_message
from plum.exceptions import PackError
from plum.utilities import pack


class TestPack:

    """Test pack transform exceptions."""

    def test_cannot_convert(self):
        """Test exception when converting to decimal."""
        with pytest.raises(PackError) as trap:
            pack("foo", DecimalX(2, 1, "big", signed=False, name="u16p1"))

        expected = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            |        |       |       | u16p1  |
            +--------+-------+-------+--------+

            ValueError occurred during pack operation:

            Value 'foo' cannot be converted to decimal.
            """
        )

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test basic features of bitfield_member() definition."""

import pytest
from enum import IntEnum

from baseline import Baseline

from plum.conformance import wrap_message
from plum.enum import EnumX
from plum.structure import Structure, bitfield_member


class TestMiscFeatures:

    """Test misc. features."""

    class Struct(Structure):
        f1 = bitfield_member(size=1)

    def test_repr(self):

        expected = Baseline(
            """
            BitFieldMember(name='f1')
            """
        )

        assert repr(self.Struct.f1) == expected

    def test_default_type(self):
        # no arg, no annotation
        assert self.Struct.f1.typ is int


class TestInvalidPosition:

    """Test invalid bit position for bit field member."""

    def test_less_than_zero(self):
        """Test bit position less than zero."""

        expected = Baseline(
            """
            bit field position must be greater than or equal to zero
            """
        )

        with pytest.raises(TypeError) as trap:
            bitfield_member(lsb=-1, size=2, signed=True)

        assert wrap_message(trap.value) == expected


class TestInvalidTypExplicit:

    """Test invalid bit field typ when specified incorrectly in member."""

    def test_instance(self):
        """Test typ is not a class but an instance of a class."""

        expected = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:
            bitfield_member(typ=1, lsb=0, size=2, signed=True)

        assert wrap_message(trap.value) == expected

    def test_invalid_class(self):
        """Test typ is not int-like."""

        expected = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:
            bitfield_member(typ=float, lsb=0, size=2, signed=True)

        assert wrap_message(trap.value) == expected


class TestInvalidTypAnnotation:

    """Test invalid bit field typ when specified via annotation."""

    def test_instance(self):
        """Test typ is not a class but an instance of a class."""
        expected = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:

            class Struct(Structure):  # pylint: disable=unused-variable
                f1 = bitfield_member(lsb=0, size=2, signed=True, typ=1)

        assert wrap_message(trap.value) == expected

    def test_invalid_class(self):
        """Test typ is not int-like."""
        expected = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:

            class Struct(Structure):  # pylint: disable=unused-variable
                f1 = bitfield_member(lsb=0, size=2, signed=True, typ=float)

        assert wrap_message(trap.value) == expected


class TestInvalidSizes:

    """Test invalid bit sizes for bit field member."""

    def test_signed_int(self):
        """Test bit size restriction for signed int."""

        # no exception
        bitfield_member(lsb=0, size=2, signed=True)

        expected = Baseline(
            """
            'size' must be 2 or greater for signed bit field
            """
        )

        with pytest.raises(ValueError) as trap:
            bitfield_member(lsb=0, size=1, signed=True)
        assert wrap_message(trap.value) == expected

        with pytest.raises(ValueError) as trap:
            bitfield_member(lsb=0, size=0, signed=True)
        assert wrap_message(trap.value) == expected

        with pytest.raises(ValueError) as trap:
            bitfield_member(lsb=0, size=-1, signed=True)
        assert wrap_message(trap.value) == expected

    def test_unsigned_int(self):
        """Test bit size restriction for unsigned int."""

        # no exception
        bitfield_member(lsb=0, size=1, signed=False)

        expected = Baseline(
            """
            'size' must be 1 or greater for unsigned bit field
            """
        )

        with pytest.raises(ValueError) as trap:
            bitfield_member(lsb=0, size=0, signed=False)
        assert wrap_message(trap.value) == expected

        with pytest.raises(ValueError) as trap:
            bitfield_member(lsb=0, size=-1, signed=False)
        assert wrap_message(trap.value) == expected


class TestInvalidInitValues:

    """Test out of range values passed to constructor."""

    class Struct(Structure):
        f1 = bitfield_member(size=4, default=0)
        f2 = bitfield_member(size=4, signed=True, default=0)

    unsigned_message = Baseline(
        """
        'f1' out of range, 0 <= f1 <= 15
        """
    )

    signed_message = Baseline(
        """
        'f2' out of range, -8 <= f2 <= 7
        """
    )

    def test_too_small_for_signed(self):
        with pytest.raises(ValueError) as trap:
            self.Struct(f2=-9)

        assert wrap_message(trap.value) == self.signed_message

    def test_too_small_for_unsigned(self):
        with pytest.raises(ValueError) as trap:
            self.Struct(f1=-1)

        assert wrap_message(trap.value) == self.unsigned_message

    def test_too_big_for_signed(self):
        with pytest.raises(ValueError) as trap:
            self.Struct(f2=8)

        assert wrap_message(trap.value) == self.signed_message

    def test_too_big_for_unsigned(self):
        with pytest.raises(ValueError) as trap:
            self.Struct(f1=16)

        assert wrap_message(trap.value) == self.unsigned_message


class TestInvalidAttributeValues:

    """Test out of range attribute values."""

    class Struct(Structure):
        f1 = bitfield_member(size=4, default=0)
        f2 = bitfield_member(size=4, signed=True, default=0)

    unsigned_message = Baseline(
        """
        out of range, 0 <= value <= 15
        """
    )

    signed_message = Baseline(
        """
        out of range, -8 <= value <= 7
        """
    )

    def test_too_small_for_signed(self):
        with pytest.raises(ValueError) as trap:
            self.Struct().f2 = -9  # pylint: disable=invalid-name

        assert wrap_message(trap.value) == self.signed_message

    def test_too_small_for_unsigned(self):
        with pytest.raises(ValueError) as trap:
            self.Struct().f1 = -1  # pylint: disable=invalid-name

        assert wrap_message(trap.value) == self.unsigned_message

    def test_too_big_for_signed(self):
        with pytest.raises(ValueError) as trap:
            self.Struct().f2 = 8

        assert wrap_message(trap.value) == self.signed_message

    def test_too_big_for_unsigned(self):
        with pytest.raises(ValueError) as trap:
            self.Struct().f1 = 16

        assert wrap_message(trap.value) == self.unsigned_message


class TestDefaultFactory:
    class Struct(Structure):
        def get_default(self):
            return self.f1 * 2

        f1 = bitfield_member(size=4)
        f2 = bitfield_member(size=4, default=get_default)

    def test_factory(self):
        assert self.Struct(f1=1).f2 == 2


class Register(IntEnum):

    PC = 0
    SP = 1
    R0 = 2
    R1 = 3


register = EnumX(name="register", enum=Register, strict=False)


class TestTolerantEnum:
    class Sample(Structure):

        m0: int = bitfield_member(lsb=4, size=4)
        m1: Register = bitfield_member(lsb=0, size=4, typ=register)

    expected_dump = Baseline(
        """
        +--------+--------+-------+----------+--------------------+
        | Offset | Access | Value | Bytes    | Format             |
        +--------+--------+-------+----------+--------------------+
        |        |        |       |          | Sample (Structure) |
        | 0      |        | 31    | 1f       |                    |
        |  [4:8] | m0     | 1     | 0001.... | int                |
        |  [0:4] | m1     | 15    | ....1111 | register           |
        +--------+--------+-------+----------+--------------------+
        """
    )

    def test_pack(self):
        _buffer, dump = self.Sample(m0=1, m1=15).ipack_and_dump()
        assert str(dump) == self.expected_dump

    def test_unpack(self):
        sample, dump = self.Sample.unpack_and_dump(b"\x1f")
        assert str(dump) == self.expected_dump
        assert sample.m0 == 1
        assert sample.m1 == 15



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test structure subclasses can override methods."""

from plum.littleendian import uint8
from plum.structure import Structure, member


class Sample(Structure):

    """Sample structure type with method overrides."""

    mbr1: int = member(fmt=uint8, ignore=True)

    def __init__(self, mbr1=0x88):
        """CUSTOM INIT"""
        # pylint: disable=super-init-not-called
        list.append(self, mbr1)

    @classmethod
    def __pack__(cls, value, pieces, dump=None):
        pieces.append(b"\x99")

    @classmethod
    def __unpack__(cls, buffer, offset, dump=None):
        structure = list.__new__(cls)
        structure[0:0] = (0x88,)
        return structure, offset + 1

    def asdict(self):
        """Return structure members in dictionary form.

        :returns: structure members
        :rtype: dict

        """
        return {"hello": "world"}

    def __eq__(self, other):
        return "__eq__"

    def __ne__(self, other):
        return "__ne__"

    def __getattr__(self, name):
        return "__getattr__"

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)


class TestOverrides:

    """Test structure subclasses can override methods.

    Certain methods are created automatically by the structure metaclass.
    Verify they all can be overridden with a custom method.

    """

    @staticmethod
    def test_init():
        """Test __init__() may be overridden."""
        assert Sample() == [0x88]

    @staticmethod
    def test_asdict():
        """Test asdict() may be overridden."""
        assert Sample().asdict() == {"hello": "world"}

    @staticmethod
    def test_pack():
        """Test __pack__() may be overridden."""
        assert Sample().ipack() == b"\x99"

    @staticmethod
    def test_unpack():
        """Test __unpack__() may be overridden."""
        sample = Sample.unpack(b"\x99")
        assert sample.mbr1 == 0x88

    @staticmethod
    def test_eq():
        """Test __eq__() may be overridden."""
        assert Sample().__eq__(None) == "__eq__"

    @staticmethod
    def test_ne():
        """Test __ne__() may be overridden."""
        assert Sample().__ne__(None) == "__ne__"

    @staticmethod
    def test_getattr():
        """Test __getattr__() may be overridden."""
        assert Sample().junk == "__getattr__"

    @staticmethod
    def test_setattr():
        """Test __getattr__() may be overridden."""
        sample = Sample()
        sample.junk = "junk"  # pylint: disable=attribute-defined-outside-init
        assert object.__getattribute__(sample, "junk") == "junk"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Interpret bytes as a string."""

import codecs
from typing import List, Optional, Tuple, Union

from ._getbytes import getbytes
from .dump import Record
from .exceptions import InsufficientMemoryError
from .transform import Transform


def _decode_str_bytes(buffer, decoder, zipped_char_bytes, zero_termination=False):
    reset = True
    for i, byte in enumerate(buffer):
        if reset:
            char_bytes = []
            zipped_char_bytes.append(["", char_bytes])
            reset = False

        if zero_termination and byte == 0:
            break

        char = decoder.decode(buffer[i : i + 1])
        char_bytes.append(byte)

        if char:
            zipped_char_bytes[-1][0] = char
            reset = True

    # FUTURE - test this
    # raise exception if bytes remain in decoder (character not complete)
    decoder.decode(b"", final=True)


def _encode_str_bytes(string, encoder, zipped_char_bytes):
    for char in string:
        zipped_char_bytes.append((char, encoder.encode(char)))

    assert not encoder.encode("", final=True)


def _iter_str_rows(zipped_char_bytes):
    row_index = 0
    row_chars, row_bytes = [], []

    for char, char_bytes in zipped_char_bytes:
        if len(char_bytes) + len(row_bytes) > 16:
            yield row_index, "".join(row_chars), bytearray(row_bytes)
            row_index += len(row_chars)
            row_chars, row_bytes = [], []

        row_chars.append(char)
        row_bytes += char_bytes

    if row_bytes:
        yield row_index, "".join(row_chars), bytearray(row_bytes)
    else:
        yield row_index, "", bytearray()


def _add_str_rows_to_dump(dump, zipped_char_bytes, buffer=None, string=None):
    nbytes = 0
    nchar = 0
    for index, chars, bytes_ in _iter_str_rows(zipped_char_bytes):
        dump.add_record(
            access=f"[{index}:{index + len(chars)}]", value=repr(chars), memory=bytes_
        )
        nbytes += len(bytes_)
        nchar += len(chars)

    if buffer:
        dump.add_extra_bytes("--error--", buffer[nbytes:])

    if string:
        string = string[nchar:]
        access = "--error--"
        for i in range(0, len(string), 16):
            dump.add_record(access).value = repr(string[i : i + 16])
            access = ""


def _add_str_bytes_to_dump(dump, buffer, decoder):
    zipped_char_bytes = []
    try:
        _decode_str_bytes(buffer, decoder, zipped_char_bytes)
    finally:
        _add_str_rows_to_dump(dump, zipped_char_bytes, buffer=buffer)


def _add_str_value_to_dump(dump, string, encoder):
    zipped_char_bytes = []
    try:
        _encode_str_bytes(string, encoder, zipped_char_bytes)
    finally:
        _add_str_rows_to_dump(dump, zipped_char_bytes, string=string)


class StrX(Transform):

    """String to bytes and bytes to string transform."""

    def __init__(
        self,
        encoding: str,
        errors: str = "strict",
        nbytes: Union[int, None] = None,
        pad: bytes = b"",
        zero_termination: bool = False,
        name: Optional[str] = None,
    ) -> None:
        # pylint: disable=too-many-arguments
        assert len(pad) <= 1

        if name is None:
            name = f"str ({encoding})"

        super().__init__(name, hint="str")

        codecs_info = codecs.lookup(encoding)

        self.__nbytes__ = nbytes
        self.__cenpz__ = codecs_info, errors, nbytes, pad, zero_termination

    @property
    def encoding(self) -> str:
        """Codecs encoding name."""
        codecs_info = self.__cenpz__[0]
        return codecs_info.name

    @property
    def errors(self) -> str:
        """Codecs error handling."""
        return self.__cenpz__[1]

    @property
    def pad(self) -> bytes:
        """Pad byte."""
        return self.__cenpz__[3]

    @property
    def zero_termination(self) -> bool:
        """Zero termination byte present."""
        return self.__cenpz__[4]

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[str, int]:
        # pylint: disable=too-many-branches,too-many-locals
        if dump is not None:
            return self.__unpack_and_dump__(buffer, offset, dump)

        codecs_info, errors, nbytes, pad, zero_termination = self.__cenpz__

        original_offset = offset

        if nbytes is None:
            chunk = buffer[offset:]
            offset += len(chunk)
        else:
            start = offset
            offset += nbytes
            chunk = buffer[start:offset]

            if len(chunk) < nbytes:
                raise InsufficientMemoryError(
                    "insufficient bytes for String, retry with dump"
                )

        if zero_termination:

            decoder = codecs_info.incrementaldecoder(errors)

            zipped_char_bytes: List[Tuple[str, bytes]] = []
            _decode_str_bytes(chunk, decoder, zipped_char_bytes, zero_termination=True)

            nstr_membytes = len(b"".join(bytes(b) for c, b in zipped_char_bytes))
            termination = chunk[nstr_membytes : nstr_membytes + 1]

            if termination != b"\x00":
                termination = b""

            if nbytes is None:
                offset = original_offset + nstr_membytes + len(termination)

            if termination != b"\x00":
                raise InsufficientMemoryError("no zero termination present")

            value = "".join(c for c, b in zipped_char_bytes)

        else:
            if pad and pad in chunk:
                chunk = chunk[: chunk.index(pad)]

            value = str(chunk, codecs_info.name, errors)

        return value, offset

    def __unpack_and_dump__(
        self, buffer: bytes, offset: int, dump: Record
    ) -> Tuple[str, int]:
        # pylint: disable=too-many-branches,too-many-locals
        codecs_info, errors, nbytes, pad, zero_termination = self.__cenpz__

        original_offset = offset

        chunk, offset = getbytes(buffer, offset, dump, nbytes)

        if zero_termination:

            decoder = codecs_info.incrementaldecoder(errors)

            dump.memory = b""

            zipped_char_bytes: List[Tuple[str, bytes]] = []
            try:
                _decode_str_bytes(
                    chunk, decoder, zipped_char_bytes, zero_termination=True
                )
            except UnicodeDecodeError:
                _add_str_rows_to_dump(dump, zipped_char_bytes, buffer=chunk)
                raise

            nstr_membytes = len(b"".join(bytes(b) for c, b in zipped_char_bytes))
            termination = chunk[nstr_membytes : nstr_membytes + 1]

            if termination == b"\x00":
                leftover_bytes = chunk[nstr_membytes + 1 :]
            else:
                leftover_bytes = chunk[nstr_membytes:]
                termination = b""

            if nbytes is None:
                offset = original_offset + nstr_membytes + len(termination)
                leftover_bytes = b""

            _add_str_rows_to_dump(dump, zipped_char_bytes)
            if termination:
                dump.add_record(access="--termination--", memory=bytes(termination))
            if leftover_bytes:
                dump.add_extra_bytes("--pad--", bytes(leftover_bytes))

            if termination != b"\x00":
                raise InsufficientMemoryError("no zero termination present")

            value = "".join(c for c, b in zipped_char_bytes)

        else:
            if pad and pad in chunk:
                i = chunk.index(pad)
                pad_bytes = chunk[i:]
                chunk = chunk[:i]
            else:
                pad_bytes = b""

            dump.memory = b""
            _add_str_bytes_to_dump(dump, chunk, codecs_info.incrementaldecoder(errors))

            if pad_bytes:
                dump.add_extra_bytes("--pad--", bytes(pad_bytes))

            value = str(chunk, codecs_info.name, errors)

        return value, offset

    def __pack__(
        self, value: str, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        if dump is not None:
            self.__pack_and_dump__(value, pieces, dump)

        else:
            codecs_info, errors, nbytes, pad, zero_termination = self.__cenpz__

            chunk = value.encode(codecs_info.name, errors)

            if zero_termination:
                chunk = chunk + b"\x00"

            actual_nbytes = len(chunk)

            if nbytes is not None:

                if actual_nbytes > nbytes:
                    raise TypeError(
                        f"number of string bytes ({actual_nbytes}) "
                        f"exceeds limit ({nbytes}) for {self!r}"
                    )

                if actual_nbytes < nbytes:
                    if not pad:
                        raise ValueError(
                            f"number of string bytes ({actual_nbytes}) "
                            f"falls short of ({nbytes}) for {self!r}"
                        )

                    chunk += pad * (nbytes - actual_nbytes)

            pieces.append(chunk)

    def __pack_and_dump__(self, value: str, pieces: List[bytes], dump: Record) -> None:
        codecs_info, errors, nbytes, pad, zero_termination = self.__cenpz__

        _add_str_value_to_dump(dump, value, codecs_info.incrementalencoder(errors))
        if zero_termination:
            dump.add_record("--termination--", memory=b"\x00")

        chunk = value.encode(codecs_info.name, errors)

        if zero_termination:
            chunk = chunk + b"\x00"

        actual_nbytes = len(chunk)

        if nbytes is not None:

            if actual_nbytes > nbytes:
                raise TypeError(
                    f"number of string bytes ({actual_nbytes}) "
                    f"exceeds limit ({nbytes}) for {self!r}"
                )

            if actual_nbytes < nbytes:
                if not pad:
                    raise ValueError(
                        f"number of string bytes ({actual_nbytes}) "
                        f"falls short of ({nbytes}) for {self!r}"
                    )
                pad = pad * (nbytes - actual_nbytes)
                chunk += pad

                dump.add_extra_bytes("--pad--", pad)

        pieces.append(chunk)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test IPv4AddressX transform properties."""

from plum.object import ObjectX


def pack(value) -> bytes:
    # pylint: disable=unused-argument
    return b""


def unpack(buffer: bytes) -> bytes:
    # pylint: disable=unused-argument
    return b""


class TestDefault:

    """Test explicitly defined with positional argument."""

    xform = ObjectX(pack, unpack, 1)

    def test_hint(self):
        assert self.xform.__hint__ == "Any"

    def test_name(self):
        assert self.xform.name == "object"

    def test_nbytes(self):
        assert self.xform.nbytes == 1


class TestPositional:

    """Test explicitly defined with positional argument."""

    xform = ObjectX(pack, unpack, 1, "name")

    def test_hint(self):
        assert self.xform.__hint__ == "Any"

    def test_name(self):
        assert self.xform.name == "name"

    def test_nbytes(self):
        assert self.xform.nbytes == 1


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    xform = ObjectX(pack=pack, unpack=unpack, nbytes=1, name="name")
